#
# naxis.py -- mode for navigating data cubes
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.ImageView import ImageViewBase
from ginga.modes.mode_base import Mode


class NaxisMode(Mode):
    """
    Naxis Mode enables bindings that can move through the slices in an
    image stack in a Ginga image viewer.

    Enter the mode by
    -----------------
    * Space, then "n"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * scroll : select previous or next slice of current axis
    * Ctrl + scroll : select previous or next axis as current axis
    * left drag : select slice as a function of percentage of cursor/window width
    * up/down arrow : select previous or next axis as current axis

    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_naxis=['__n', None, None],

            kp_naxis_up=['naxis+up', 'naxis+b'],
            kp_naxis_dn=['naxis+down', 'naxis+n'],
            sc_naxis_slice=['naxis+scroll'],
            sc_naxis_axis=['naxis+ctrl+scroll'],
            ms_naxis=['naxis+left'],
            pa_naxis=['naxis+pan'])

        # the axis being examined
        self.axis = 3

    def __str__(self):
        return 'naxis'

    def start(self):
        self.axis = 3

    def stop(self):
        self.onscreen_message(None)

    def _nav_naxis(self, viewer, axis, direction, msg=True):
        """Interactively change the slice of the image in a data cube
        by scrolling.
        """
        image = viewer.get_image()
        if image is None:
            self.onscreen_message("No image", delay=1.0)
            return
        if axis < 3 or axis > len(image.axisdim):
            # attempting to access a non-existent axis
            self.onscreen_message("Bad axis: %d" % (self.axis), delay=1.0)
            return
        _axis = len(image.axisdim) - axis

        axis_lim = image.axisdim[_axis]
        naxispath = list(image.naxispath)
        m = axis - 3

        idx = naxispath[m]
        if direction == 'down':
            idx = (idx + 1) % axis_lim
        elif direction == 'up':
            idx = idx - 1
            if idx < 0:
                idx = axis_lim - 1
        else:
            # no change, except possibly axis
            pass

        naxispath[m] = idx
        image.set_naxispath(naxispath)
        if msg:
            self.onscreen_message("axis: %d  slice: %d" % (axis, idx),
                                  delay=1.0)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_naxis_up(self, viewer, event, data_x, data_y, msg=True):
        event.accept()
        self.axis += 1
        self._nav_naxis(viewer, self.axis, 'same', msg=msg)

    def kp_naxis_dn(self, viewer, event, data_x, data_y, msg=True):
        event.accept()
        self.axis -= 1
        self._nav_naxis(viewer, self.axis, 'same', msg=msg)

    #####  SCROLL ACTION CALLBACKS #####

    def sc_naxis_slice(self, viewer, event, msg=True):
        """Interactively change the slice of the image in a data cube
        by scrolling.
        """
        event.accept()
        direction = self.get_direction(event.direction)

        self._nav_naxis(viewer, self.axis, direction, msg=msg)

    def sc_naxis_axis(self, viewer, event, msg=True):
        """Interactively change the slice of the image in a data cube
        by scrolling.
        """
        event.accept()
        direction = self.get_direction(event.direction)
        if direction == 'up':
            self.axis += 1
        elif direction == 'down':
            self.axis -= 1

        self._nav_naxis(viewer, self.axis, 'same', msg=msg)

    #####  MOUSE ACTION CALLBACKS #####

    def ms_naxis(self, viewer, event, data_x, data_y, msg=True):

        event.accept()
        x, y = self.get_win_xy(viewer)

        image = viewer.get_image()
        if image is None:
            self.onscreen_message("No image", delay=1.0)
            return

        # which axis (in FITS NAXIS terminology)
        if self.axis < 3 or self.axis > len(image.axisdim):
            # attempting to access a non-existant axis
            self.onscreen_message("Bad axis: %d" % (self.axis), delay=1.0)
            return

        _axis = len(image.axisdim) - self.axis
        axis_lim = image.axisdim[_axis]
        naxispath = list(image.naxispath)
        m = self.axis - 3

        if event.state in ('down', 'move'):
            win_wd, win_ht = viewer.get_window_size()
            x_pct = min(max(0.0, x / float(win_wd)), 1.0)
            idx = int(round(x_pct * (axis_lim - 1)))
            naxispath[m] = idx
            image.set_naxispath(naxispath)
            if msg:
                self.onscreen_message("axis: %d  slice: %d" % (self.axis, idx),
                                      delay=1.0)

    ##### GESTURE ACTION CALLBACKS #####

    def pa_naxis(self, viewer, event, msg=True):
        """Interactively change the slice of the image in a data cube
        by pan gesture.
        """
        event.accept()
        event = self._pa_synth_scroll_event(event)
        if event.state != 'move':
            return

        # TODO: be able to pick axis
        direction = self.get_direction(event.direction)

        self._nav_naxis(viewer, self.axis, direction, msg=msg)



#
# events.py -- Event classes for Ginga viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#


class UIEvent:
    """Base class for user interface events."""
    def __init__(self, viewer=None):
        self.viewer = viewer
        self.handled = False

    def accept(self):
        self.handled = True

    def was_handled(self):
        return self.handled


class KeyEvent(UIEvent):
    """A key press or release event in a Ginga viewer.

    Attributes
    ----------
    key : str
        The key as it is known to Ginga

    state: str
        'down' if a key press, 'up' if a key release

    mode : str
        The mode name of the mode that was active when the event happened

    modifiers : set of str
        A set of names of modifier keys that were pressed at the time

    data_x : float
        X part of the data coordinates of the viewer under the cursor

    data_y : float
        Y part of the data coordinates of the viewer under the cursor

    viewer : subclass of `~ginga.ImageView.ImageViewBase`
        The viewer in which the event happened
    """
    def __init__(self, key=None, state=None, mode=None, modifiers=None,
                 data_x=None, data_y=None, viewer=None):
        super().__init__(viewer=viewer)
        self.key = key
        self.state = state
        self.mode = mode
        self.modifiers = modifiers
        self.data_x = data_x
        self.data_y = data_y


class PointEvent(UIEvent):
    """A mouse/pointer/cursor event in a Ginga viewer.

    Attributes
    ----------
    button : str
        The name of the button as set up in the configuration

    state: str
        'down' if a press, 'move' if being dragged, 'up' if a release

    mode : str
        The mode name of the mode that was active when the event happened

    modifiers : set of str
        A set of names of modifier keys that were pressed at the time

    data_x : float
        X part of the data coordinates of the viewer under the cursor

    data_y : float
        Y part of the data coordinates of the viewer under the cursor

    viewer : subclass of `~ginga.ImageView.ImageViewBase`
        The viewer in which the event happened
    """
    def __init__(self, button=None, state=None, mode=None, modifiers=None,
                 data_x=None, data_y=None, viewer=None):
        super().__init__(viewer=viewer)
        self.button = button
        self.state = state
        self.mode = mode
        self.modifiers = modifiers
        self.data_x = data_x
        self.data_y = data_y


class ScrollEvent(UIEvent):
    """A mouse or trackpad scroll event in a Ginga viewer.

    Attributes
    ----------
    button : str
        The name of the button as set up in the configuration

    state: str
        Always 'scroll'

    mode : str
        The mode name of the mode that was active when the event happened

    modifiers : set of str
        A set of names of modifier keys that were pressed at the time

    direction : float
        A direction in compass degrees of the scroll

    amount : float
        The amount of the scroll

    data_x : float
        X part of the data coordinates of the viewer under the cursor

    data_y : float
        Y part of the data coordinates of the viewer under the cursor

    viewer : subclass of `~ginga.ImageView.ImageViewBase`
        The viewer in which the event happened
    """
    def __init__(self, button=None, state=None, mode=None, modifiers=None,
                 direction=None, amount=None, data_x=None, data_y=None,
                 viewer=None):
        super().__init__(viewer=viewer)
        self.button = button
        self.state = state
        self.mode = mode
        self.modifiers = modifiers
        self.direction = direction
        self.amount = amount
        self.data_x = data_x
        self.data_y = data_y


class PinchEvent(UIEvent):
    """A pinch event in a Ginga viewer.

    Attributes
    ----------
    button : str
        The name of the button as set up in the configuration

    state: str
        'start' (gesture starting), 'move' (in action) or 'stop' (done)

    mode : str
        The mode name of the mode that was active when the event happened

    modifiers : set of str
        A set of names of modifier keys that were pressed at the time

    rot_deg : float
        Amount of rotation in degrees

    scale : float
        Scale of the pinch shrink or enlargement

    data_x : float
        X part of the data coordinates of the viewer under the cursor

    data_y : float
        Y part of the data coordinates of the viewer under the cursor

    viewer : subclass of `~ginga.ImageView.ImageViewBase`
        The viewer in which the event happened
    """
    def __init__(self, button=None, state=None, mode=None, modifiers=None,
                 rot_deg=None, scale=None, data_x=None, data_y=None,
                 viewer=None):
        super().__init__(viewer=viewer)
        self.button = button
        self.state = state
        self.mode = mode
        self.modifiers = modifiers
        self.rot_deg = rot_deg
        self.scale = scale
        self.data_x = data_x
        self.data_y = data_y


class PanEvent(UIEvent):
    """A pinch event in a Ginga viewer.

    Attributes
    ----------
    button : str
        The name of the button as set up in the configuration

    state: str
        'start' (gesture starting), 'move' (in action) or 'stop' (done)

    mode : str
        The mode name of the mode that was active when the event happened

    modifiers : set of str
        A set of names of modifier keys that were pressed at the time

    delta_x : float
        Amount of scroll movement in the X direction

    delta_y : float
        Amount of scroll movement in the Y direction

    data_x : float
        X part of the data coordinates of the viewer under the cursor

    data_y : float
        Y part of the data coordinates of the viewer under the cursor

    viewer : subclass of `~ginga.ImageView.ImageViewBase`
        The viewer in which the event happened
    """
    def __init__(self, button=None, state=None, mode=None, modifiers=None,
                 delta_x=None, delta_y=None, data_x=None, data_y=None,
                 viewer=None):
        super().__init__(viewer=viewer)
        self.button = button
        self.state = state
        self.mode = mode
        self.modifiers = modifiers
        self.delta_x = delta_x
        self.delta_y = delta_y
        self.data_x = data_x
        self.data_y = data_y



#
# io_rgb.py -- RGB image file handling.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time
import mimetypes
from io import BytesIO

import numpy as np
from PIL import Image
from PIL.ExifTags import TAGS

from ginga.BaseImage import Header, ImageError
from ginga.util import iohelper, rgb_cms
from ginga.util.io import io_base
from ginga.misc import Bunch
from ginga import trcalc

# optional imports
try:
    # do we have opencv available?
    import cv2
    have_opencv = True
except ImportError:
    have_opencv = False

# exifread library for getting metadata, in case we are using OpenCv
try:
    import exifread
    have_exif = True
except ImportError:
    have_exif = False

try:
    # for opening HEIC (Apple image) files)
    from pillow_heif import register_heif_opener
    register_heif_opener()
except ImportError:
    pass
# For testing...
#have_exif = False
#have_opencv = False


def load_file(filepath, idx=None, logger=None, **kwargs):
    """
    Load an object from a RGB file.

    """
    opener = RGBFileHandler(logger)
    return opener.load_file(filepath, **kwargs)


class BaseRGBFileHandler(io_base.BaseIOHandler):

    name = 'RGB'

    def __init__(self, logger):
        super(BaseRGBFileHandler, self).__init__(logger)

        self._path = None
        self.fileinfo = None

        self.clr_mgr = rgb_cms.ColorManager(self.logger)

    def load_file(self, filespec, dstobj=None, **kwargs):
        info = iohelper.get_fileinfo(filespec)
        if not info.ondisk:
            raise ValueError("File does not appear to be on disk: %s" % (
                info.url))

        filepath = info.filepath
        if dstobj is None:
            # Put here to avoid circular import
            from ginga.RGBImage import RGBImage
            dstobj = RGBImage(logger=self.logger)

        header = Header()
        metadata = {'header': header, 'path': filepath}

        data_np = self.imload(filepath, metadata)

        dstobj.set_data(data_np, metadata=metadata)

        if dstobj.name is not None:
            dstobj.set(name=dstobj.name)
        else:
            name = iohelper.name_image_from_path(filepath, idx=None)
            dstobj.set(name=name)

        if 'order' in metadata:
            dstobj.order = metadata['order']
        dstobj.set(path=filepath, idx=None, image_loader=self.load_file)
        return dstobj

    def open_file(self, filespec, **kwargs):
        self._path = filespec
        return self

    def close(self):
        self._path = None

    def __len__(self):
        return 1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def __del__(self):
        self.close()

    def load_idx_cont(self, idx_spec, loader_cont_fn, **kwargs):

        # TODO: raise an error if idx_spec doesn't match a single image
        idx = 0
        if idx_spec is not None and idx_spec != '':
            idx = int(idx_spec)

        data_obj = self.load_idx(idx, **kwargs)

        # call continuation function
        loader_cont_fn(data_obj)

    def imload(self, filepath, metadata):
        """Load an image file, guessing the format, and return a numpy
        array containing an RGB image.  If EXIF keywords can be read
        they are returned in the metadata.
        """
        start_time = time.time()
        typ, enc = mimetypes.guess_type(filepath)
        if not typ:
            typ = 'image/jpeg'
        typ, subtyp = typ.split('/')
        self.logger.debug("MIME type is %s/%s" % (typ, subtyp))

        data_np = self._imload(filepath, metadata)

        end_time = time.time()
        self.logger.debug("loading time %.4f sec" % (end_time - start_time))
        return data_np

    def imresize(self, data, new_wd, new_ht, method='bilinear'):
        """Scale an image in numpy array _data_ to the specified width and
        height.  A smooth scaling is preferred.
        """
        start_time = time.time()

        newdata = self._imresize(data, new_wd, new_ht, method=method)

        end_time = time.time()
        self.logger.debug("scaling time %.4f sec" % (end_time - start_time))

        return newdata

    def get_thumb(self, filepath):
        if not have_exif:
            raise Exception("Install exifread to use this method")

        try:
            with open(filepath, 'rb') as in_f:
                info = exifread.process_file(in_f)
            buf = info['JPEGThumbnail']

        except Exception as e:
            return None

        image = Image.open(BytesIO(buf))
        data_np = np.array(image)
        return data_np

    def _getexif(self, filepath, kwds):
        if have_exif:
            try:
                with open(filepath, 'rb') as in_f:
                    info = exifread.process_file(in_f, details=False)
                if info is not None:
                    kwds.update(info)
                    o_tag = 'Image Orientation'
                    if o_tag in info:
                        val = info[o_tag].values[0]
                        kwds[o_tag] = val

            except Exception as e:
                self.logger.warning("Failed to get image metadata: %s" % (str(e)))

        else:
            self.logger.warning("Please install 'exifread' module to get image metadata")

    def get_buffer(self, data_np, header, format, output=None):
        """Get image as a buffer in (format).
        Format should be 'jpeg', 'png', etc.
        """
        image = Image.fromarray(data_np)
        buf = output
        if buf is None:
            buf = BytesIO()
        image.save(buf, format)
        return buf

    def get_frame(self, num):
        raise NotImplementedError("subclass should implement this method")

    def get_directory(self):
        return self.hdu_db

    def get_info_idx(self, idx):
        return self.hdu_db[idx]


class OpenCvFileHandler(BaseRGBFileHandler):
    """For loading common RGB image formats (e.g. JPEG, etc).
    """

    name = 'OpenCv'
    mimetypes = ['image/jpeg',
                 'image/tiff',
                 'image/png',
                 'image/gif',
                 'image/x-portable-pixmap',
                 'image/ppm',
                 'image/x-portable-anymap',
                 'image/pnm',
                 'image/x-portable-bitmap',
                 'image/pbm',
                 'image/x-portable-graymap',
                 'image/x-portable-greymap',
                 'image/pgm',
                 'image/vnd.microsoft.icon',
                 'image/bmp',
                 'image/x-tga',
                 'video/x-msvideo',
                 'video/mp4',
                 'video/quicktime',
                 'video/x-matroska']

    @classmethod
    def check_availability(cls):
        import cv2   # noqa

    def open_file(self, filespec, **kwargs):

        info = iohelper.get_fileinfo(filespec)
        if not info.ondisk:
            raise ImageError("File does not appear to be on disk: %s" % (
                info.url))

        self.fileinfo = info
        filepath = info.filepath

        self._path = filepath
        self.rgb_f = cv2.VideoCapture(filepath)

        idx = 0
        extver_db = {}
        self.hdu_info = []
        self.hdu_db = {}
        self.numframes = int(self.rgb_f.get(cv2.CAP_PROP_FRAME_COUNT))
        self.logger.debug("number of frames: {}".format(self.numframes))

        name = "image{}".format(idx)
        extver = 0
        # prepare a record of pertinent info about the HDU for
        # lookups by numerical index or (NAME, EXTVER)
        # NOTE: dtype is uint8 even though we may later read it at a higher
        # bit depth--because there doesn't seem to be a straightforward way
        # to know the bit depth before reading the data, unfortunately
        d = Bunch.Bunch(index=idx, name=name, extver=extver,
                        dtype='uint8', htype='imagehdu')
        self.hdu_info.append(d)
        # different ways of accessing this HDU:
        # by numerical index
        self.hdu_db[idx] = d
        # by (hduname, extver)
        key = (name, extver)
        if key not in self.hdu_db:
            self.hdu_db[key] = d

        self.extver_db = extver_db
        return self

    def close(self):
        self._path = None
        if self.rgb_f is not None:
            self.rgb_f.release()
        self.rgb_f = None

    def __len__(self):
        return len(self.hdu_info)

    def load_idx(self, idx, **kwargs):
        if self.rgb_f is None:
            raise ValueError("Please call open_file() first!")

        if idx is None:
            idx = 0

        metadata = dict()
        if idx == 0:
            data_np = self.imload(self.fileinfo.filepath, metadata)
        else:
            raise IndexError(f"index {idx} out of range")

        from ginga.RGBImage import RGBImage
        data_obj = RGBImage(data_np=data_np, metadata=metadata,
                            logger=self.logger)
        data_obj.io = self

        data_obj.axisdim = [self.numframes] + list(data_np.shape[:2])
        data_obj.naxispath = [0, 0, 0]

        name = self.fileinfo.name + '[{}]'.format(idx)
        data_obj.set(name=name, path=self.fileinfo.filepath, idx=idx)

        return data_obj

    def get_frame(self, num, metadata=None):
        if self.rgb_f is None:
            if self.fileinfo is None:
                raise ValueError("Please call open_file() first!")
            # re-open file/stream, if not open already
            self._path = self.fileinfo.filepath
            self.rgb_f = cv2.VideoCapture(self._path)

        if num < 0 or num >= self.numframes:
            raise IndexError(f"Frame outside valid range of 0-{self.numframes}")

        self.rgb_f.set(cv2.CAP_PROP_POS_FRAMES, num)
        okay, data_np = self.rgb_f.read()
        if not okay:
            raise IndexError(f"Error reading frame {num}")

        data_np = self._reorder_array(data_np, metadata=metadata)

        return data_np

    def save_file_as(self, filepath, data_np, header):
        # TODO: save keyword metadata!
        if not have_opencv:
            raise ImageError("Install 'opencv' to be able to save images")

        # First choice is OpenCv, because it supports high-bit depth
        # multiband images
        cv2.imwrite(filepath, data_np)

    def _imload(self, filepath, metadata):
        if not have_opencv:
            raise ImageError("Install 'opencv' to be able to load images")

        # OpenCv supports high-bit depth multiband images if you use
        # IMREAD_UNCHANGED.
        # NOTE: IMREAD_IGNORE_ORIENTATION does not seem to be obeyed here!
        # Seems to need to be combined with IMREAD_COLOR, which forces an
        # 8bpp image
        data_np = cv2.imread(filepath,
                             cv2.IMREAD_UNCHANGED + cv2.IMREAD_IGNORE_ORIENTATION)
        if data_np is None:
            self.rgb_f.set(cv2.CAP_PROP_POS_FRAMES, 0)
            okay, data_np = self.rgb_f.read()
            if not okay:
                raise IndexError("Error reading frame 0")

        return self._process_opencv_array(data_np, metadata, filepath)

    def _process_opencv_array(self, data_np, metadata, filepath):

        kwds = metadata.get('header', None)
        if kwds is None:
            kwds = Header()
            metadata['header'] = kwds

        # OpenCv doesn't "do" image metadata, so we punt to exifread
        # library (if installed)
        self._getexif(filepath, kwds)

        data_np = self._reorder_array(data_np, metadata)

        # OpenCv added a feature to do auto-orientation when loading
        # (see https://github.com/opencv/opencv/issues/4344)
        # So reset these values to prevent auto-orientation from
        # happening later
        # NOTE: this is only needed if IMREAD_IGNORE_ORIENTATION
        # (see above) is not working
        kwds['Orientation'] = 1
        kwds['Image Orientation'] = 1

        return data_np

    def _reorder_array(self, data_np, metadata=None):
        # opencv returns BGR images, whereas PIL and others return RGB
        if len(data_np.shape) >= 3 and data_np.shape[2] >= 3:
            if data_np.shape[2] == 3:
                order = 'BGR'
                dst_order = 'RGB'
            else:
                order = 'BGRA'
                dst_order = 'RGBA'
            data_np = trcalc.reorder_image(dst_order, data_np, order)

            if metadata is not None:
                metadata['order'] = dst_order

                # convert to working color profile, if can
                if self.clr_mgr.can_profile():
                    kwds = metadata.get('header', None)
                    if kwds is None:
                        kwds = Header()
                        metadata['header'] = kwds
                    data_np = self.clr_mgr.profile_to_working_numpy(data_np,
                                                                    kwds)

        return data_np

    def _imresize(self, data, new_wd, new_ht, method='bilinear'):
        # TODO: take into account the method parameter
        if not have_opencv:
            raise ImageError("Install 'opencv' to be able to resize RGB images")

        # First choice is OpenCv, because it supports high-bit depth
        # multiband images
        newdata = cv2.resize(data, dsize=(new_wd, new_ht),
                             interpolation=cv2.INTER_CUBIC)

        return newdata


class PillowFileHandler(BaseRGBFileHandler):
    """For loading common RGB image formats (e.g. JPEG, etc).
    """
    name = 'Pillow'
    mimetypes = ['image/jpeg',
                 'image/png',
                 'image/tiff',
                 'image/gif',
                 'image/x-portable-pixmap',
                 'image/ppm',
                 'image/x-portable-anymap',
                 'image/pnm',
                 'image/x-portable-bitmap',
                 'image/pbm',
                 'image/x-portable-graymap',
                 'image/x-portable-greymap',
                 'image/pgm',
                 'image/bmp',
                 'image/x-tga',
                 'image/x-icns',
                 'image/vnd.microsoft.icon',
                 'image/heif']

    @classmethod
    def check_availability(cls):
        from PIL import Image    # noqa

    def open_file(self, filespec, **kwargs):

        info = iohelper.get_fileinfo(filespec)
        if not info.ondisk:
            raise ImageError("File does not appear to be on disk: %s" % (
                info.url))

        self.fileinfo = info
        filepath = info.filepath

        self._path = filepath
        self.rgb_f = Image.open(filepath)

        idx = 0
        extver_db = {}
        self.hdu_info = []
        self.hdu_db = {}
        if hasattr(self.rgb_f, 'n_frames'):
            self.numframes = self.rgb_f.n_frames - 1
        else:
            self.numframes = 0
        self.logger.debug("number of frames: {}".format(self.numframes))

        name = "image{}".format(idx)
        extver = 0
        # prepare a record of pertinent info about the HDU for
        # lookups by numerical index or (NAME, EXTVER)
        d = Bunch.Bunch(index=idx, name=name, extver=extver,
                        dtype='uint8', htype='imagehdu')
        self.hdu_info.append(d)
        # different ways of accessing this HDU:
        # by numerical index
        self.hdu_db[idx] = d
        # by (hduname, extver)
        key = (name, extver)
        if key not in self.hdu_db:
            self.hdu_db[key] = d

        self.extver_db = extver_db
        return self

    def close(self):
        self._path = None
        self.rgb_f = None

    def __len__(self):
        return len(self.hdu_info)

    def get_frame(self, num, metadata=None):
        if self.rgb_f is None:
            if self.fileinfo is None:
                raise ValueError("Please call open_file() first!")
            # re-open file, if not open already
            self._path = self.fileinfo.filepath
            self.rgb_f = Image.open(self._path)

        if num < 0 or num >= self.numframes:
            raise IndexError(f"Frame outside valid range of 0-{self.numframes}")

        # NOTE: pillow does not seem to be consistent on supporting
        # 0-based indexing on RGB images--it seems to depend on whether
        # there is a "base image"
        #self.rgb_f.seek(num)
        self.rgb_f.seek(num + 1)

        image_pil = self.rgb_f
        return self._process_image(image_pil, metadata=metadata)

    def save_file_as(self, filepath, data_np, header):
        # TODO: save keyword metadata!
        img = Image.fromarray(data_np)

        # pillow is not happy saving images to JPG with an alpha channel
        img = img.convert('RGB')

        img.save(filepath)

    def load_idx(self, idx, **kwargs):
        if self.rgb_f is None:
            raise ValueError("Please call open_file() first!")

        if idx is None:
            idx = 0

        if idx > 0:
            raise IndexError(f"index {idx} out of range")

        kwds = Header()
        metadata = dict(header=kwds)

        if self.numframes > 0:
            self.rgb_f.seek(idx + 1)

        image_pil = self.rgb_f
        data_np = self._process_image(image_pil, metadata=metadata)

        from ginga.RGBImage import RGBImage
        data_obj = RGBImage(data_np=data_np, metadata=metadata,
                            logger=self.logger, order=image_pil.mode)
        data_obj.io = self

        data_obj.axisdim = [self.numframes] + list(data_np.shape[:2])
        data_obj.naxispath = [0, 0, 0]

        name = self.fileinfo.name + '[{}]'.format(idx)
        data_obj.set(name=name, path=self.fileinfo.filepath, idx=idx,
                     header=kwds)

        return data_obj

    def _get_header(self, image_pil, kwds):
        if hasattr(image_pil, 'getexif'):
            info = image_pil.getexif()
            if info is not None:
                for tag, value in info.items():
                    kwd = TAGS.get(tag, tag)
                    kwds[kwd] = value

        else:
            self.logger.warning("can't get EXIF data; no getexif() method")

        # is there an embedded color profile?
        if 'icc_profile' in image_pil.info:
            kwds['icc_profile'] = image_pil.info['icc_profile']

    def _imload(self, filepath, metadata):
        image_pil = Image.open(filepath)
        data_np = self._process_image(image_pil, metadata=metadata)
        return data_np

    def _process_image(self, image_pil, metadata=None):
        if metadata is not None:
            metadata['order'] = image_pil.mode
            kwds = metadata.get('header', None)
            if kwds is None:
                kwds = Header()
                metadata['header'] = kwds
        else:
            kwds = Header()

        try:
            self._get_header(image_pil, kwds)
        except Exception as e:
            self.logger.warning("Failed to get image metadata: {!r}".format(e))

        # convert to working color profile, if can
        if self.clr_mgr.can_profile():
            image_pil = self.clr_mgr.profile_to_working_pil(image_pil, kwds)

        # convert from PIL to numpy
        data_np = np.array(image_pil)
        return data_np

    def _imresize(self, data, new_wd, new_ht, method='bilinear'):
        # TODO: take into account the method parameter

        image_pil = Image.fromarray(data)
        image_pil = image_pil.resize((new_wd, new_ht), Image.BICUBIC)
        newdata = np.array(image_pil)

        return newdata


RGBFileHandler = PillowFileHandler



#
# vip.py -- data extraction operations on mixed, plotted images.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time
import numpy as np
# for debugging
#np.set_printoptions(threshold=np.inf)

from ginga import trcalc
from ginga.canvas.types.image import ImageP
from ginga.canvas.types.layer import Canvas
from ginga.misc import Bunch


class ViewerImageProxy:
    """This class can be used in lieu of a `~ginga.BaseImage`-subclassed
    object (such as `~ginga.AstroImage` to handle cases where either an
    image has been loaded into a viewer in the conventional way, OR multiple
    images have been plotted onto the viewer canvas.  It has a subset of the
    API for an image wrapper object, and can be substituted for that in cases
    that only use the subset.

    Every `ImageView`-based viewer has one already constructed inside it.

    Example. Previously, one might do:

       image = viewer.get_image()
       data, x1, y1, x2, y2 = image.cutout_radius(data_x, data_y, radius)

    Alternative, using the already built vip:

        image = viewer.get_vip()
        data, x1, y1, x2, y2 = image.cutout_radius(data_x, data_y, radius)

    Here are the methods that are supported:
    * cutout_data
    * cutout_adjust
    * cutout_radius
    * cutout_cross
    * get_pixels_on_line
    * get_pixels_on_curve
    * get_shape_view
    * cutout_shape
    * get_size
    * get_depth
    * get_center
    * get_minmax
    * get_data_xy
    * info_xy
    * pixtoradec
    * has_valid_wcs
    * _slice

    Supported properties:
    * shape
    * width
    * height
    * depth
    * ndim
    """

    def __init__(self, viewer):
        """Constructor for a ViewerImageProxy object.

        Parameters
        ----------
        viewer : `~ginga.ImageView` (or subclass thereof)
            Ginga viewer object

        """
        self.viewer = viewer
        self.logger = viewer.get_logger()

        self.limit_cutout = 5000

    def get_canvas_images_at_pt(self, pt):
        """Extract the canvas Image-based objects under the point.

        Parameters
        ----------
        pt : tuple of int
            Point in data coordinates; e.g. (data_x, data_y)

        Returns
        -------
        obj : `~ginga.canvas.types.image.Image` (or subclass thereof)
            The top canvas image object found under this point

        """
        # get first canvas image object found under the cursor
        data_x, data_y = pt[:2]
        canvas = self.viewer.get_canvas()
        objs = canvas.get_items_at(pt)
        objs = list(filter(lambda obj: isinstance(obj, ImageP), objs))
        # top most objects are farther down
        objs.reverse()
        return objs

    def get_image_at_pt(self, pt):
        """Extract the image wrapper object under the point.

        Parameters
        ----------
        pt : tuple of int
            Point in data coordinates; e.g. (data_x, data_y)

        Returns
        -------
        (image, pt2) : tuple
            `image` is a `~ginga.BaseImage.BaseImage` (or subclass thereof)
            and `pt2` is a modified version of `pt` with the plotted location
            subtracted.

        """
        objs = self.get_canvas_images_at_pt(pt)
        data_x, data_y = pt[:2]
        off = self.viewer.data_off

        for obj in objs:
            if not obj.is_data:
                continue
            image = obj.get_image()
            # adjust data coords for where this image is plotted
            _x, _y = data_x - obj.x, data_y - obj.y
            order = image.get_order()
            if 'A' not in order:
                # no alpha channel, so this image's data is valid
                return (image, (_x, _y))

            aix = order.index('A')
            data = image.get_data()
            _d_x, _d_y = int(np.floor(_x + off)), int(np.floor(_y + off))
            val = data[_d_y, _d_x]
            if np.isclose(val[aix], 0.0):
                # alpha value is 0
                continue

            return (image, (_x, _y))

        return None, pt

    def getval_pt(self, pt):
        """Extract the data value from an image under the point.

        The value will be NaN if the point does not refer to a valid
        location within a plotted image.

        Parameters
        ----------
        pt : tuple of int
            Point in data coordinates; e.g. (data_x, data_y)

        Returns
        -------
        val : `numpy` value
            The value for the image object found under this point

        """
        val = self.get_data_xy(pt[0], pt[1])
        if val is None:
            return np.nan
        return val

    ## def extend_view(self, image, view):
    ##     if len(image.shape) <= 2:
    ##         return view
    ##     order = image.get_order()
    ##     if 'M' in order:
    ##         idx = order.index('M')
    ##     else:
    ##         idx = 0
    ##     return view + (idx:idx+1,)

    def get_images(self, res, canvas):
        for obj in canvas.objects:
            if isinstance(obj, ImageP) and obj.is_data:
                res.append(obj)
            elif isinstance(obj, Canvas):
                self.get_images(res, obj)
        return res

    # ----- for compatibility with BaseImage objects -----

    def cutout_data(self, x1, y1, x2, y2, xstep=1, ystep=1, z=0,
                    astype=float, fill_value=np.nan):
        """Cut out data area based on rectangular coordinates.

        Parameters
        ----------
        x1 : int
            Starting X coordinate

        y1 : int
            Starting Y coordinate

        x2 : int
            (One more than) Ending X coordinate

        y2 : int
            (one more than) Ending Y coordinate

        xstep : int (optional, default: 1)
            Step in X direction

        ystep : int (optional, default: 1)
            Step in Y direction

        z : None or int (optional, default: 0)
            Index of a Z slice, if cutout array has three dimensions

        astype : None, str or `numpy.dtype` (optional, default: None)
            Optional type to coerce the result array to

        Returns
        -------
        arr : `numpy.ndarray`
            The cut out array

        """
        t1 = time.time()

        if astype is None:
            astype = float
        if fill_value is None:
            fill_value = np.nan

        # coerce to int values, just in case
        x1, y1, x2, y2 = trcalc.sort_xy(x1, y1, x2, y2)
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        xstep, ystep = int(xstep), int(ystep)

        # output size
        x_len = len(range(x1, x2, xstep))
        y_len = len(range(y1, y2, ystep))

        # create result array, filled with fill value
        data_np = np.full((y_len, x_len), fill_value, dtype=astype)

        # calculate pixel containment indexes in cutout bbox
        yi, xi = np.mgrid[y1:y2:ystep, x1:x2:xstep]
        pts = np.asarray((xi, yi)).T

        canvas = self.viewer.get_canvas()
        images = self.get_images([], canvas)

        # iterate through images on this canvas, filling the result
        # array with pixels that overlap in each image
        for cv_img in images:
            # quick check for images overlapping our bbox;
            # skip those that do not overlap
            _x1, _y1, _x2, _y2 = cv_img.get_llur()
            dx = min(x2, _x2) - max(x1, _x1)
            dy = min(y2, _y2) - max(y1, _y1)
            if dx < 0 or dy < 0:
                continue

            # make a dst mask to assign only those pixels in this image
            # overlapping the cutout bbox.
            mask = cv_img.contains_pts(pts)
            # now form indices into this image's array to extract
            # the overlapping pixels
            xpos, ypos = cv_img.crdmap.to_data((cv_img.x, cv_img.y))
            xpos, ypos = int(xpos), int(ypos)
            xstart, ystart = x1 - xpos, y1 - ypos
            xstop, ystop = x2 - xpos, y2 - ypos
            yii, xii = np.mgrid[ystart:ystop:ystep, xstart:xstop:xstep]
            yi, xi = yii[mask].ravel(), xii[mask].ravel()

            image = cv_img.get_image()
            src_data = image.get_data()
            # check if this image has an alpha mask; if so, then we need
            # to further modify the dst mask to ignore the alpha masked pixels
            # NOTE: currently treatment of alpha channel is binary--
            # you get it or you don't; but it may be possible to do some
            # kind of blending in the future
            if len(src_data.shape) > 2:
                img_data = src_data[..., z][yi, xi]
                order = image.get_order()
                if 'A' in order:
                    ai = order.index('A')
                    a_arr = src_data[..., ai][yi, xi]
                    amask = np.logical_not(np.isclose(a_arr, 0))
                    mask[np.nonzero(mask)] = amask
                    data_np[mask] = img_data[amask]
                else:
                    data_np[mask] = img_data
            else:
                data_np[mask] = src_data[yi, xi]

        self.logger.debug("time to cutout_data %.4f sec" % (time.time() - t1))
        return data_np

    def cutout_adjust(self, x1, y1, x2, y2, xstep=1, ystep=1, z=0,
                      astype=None):
        """Like cutout_data(), but modifies the coordinates to remain within
        the boundaries of plotted data.

        Returns the cutout data slice and (possibly modified) bounding box.
        """
        dx = x2 - x1
        dy = y2 - y1

        xy_mn, xy_mx = self.viewer.get_limits()

        if x2 >= xy_mx[0]:
            x2 = xy_mx[0]
            x1 = x2 - dx

        if x1 < xy_mn[0]:
            x1 = xy_mn[0]

        if y2 >= xy_mx[1]:
            y2 = xy_mx[1]
            y1 = y2 - dy

        if y1 < xy_mn[1]:
            y1 = xy_mn[1]

        data = self.cutout_data(x1, y1, x2, y2, xstep=xstep, ystep=ystep,
                                z=z, astype=astype)
        return (data, x1, y1, x2, y2)

    def cutout_radius(self, x, y, radius, xstep=1, ystep=1, astype=None):
        return self.cutout_adjust(x - radius, y - radius,
                                  x + radius + 1, y + radius + 1,
                                  xstep=xstep, ystep=ystep, z=0,
                                  astype=astype)

    def cutout_cross(self, x, y, radius):
        """Cut two data subarrays that have a center at (x, y) and with
        radius (radius) from (image).  Returns the starting pixel (x0, y0)
        of each cut and the respective arrays (xarr, yarr).
        """
        x, y, radius = int(x), int(y), int(radius)
        x1, x2 = x - radius, x + radius
        y1, y2 = y - radius, y + radius
        data_np = self.cutout_data(x1, y1, x2 + 1, y2 + 1)

        xarr = data_np[y - y1, :]
        yarr = data_np[:, x - x1]

        return (x1, y1, xarr, yarr)

    ## def get_shape_mask(self, shape_obj):
    ##     """
    ##     Return full mask where True marks pixels within the given shape.
    ##     """
    ##     xy_mn, xy_mx = self.viewer.get_limits()
    ##     yi = np.mgrid[xy_mn[1]:xy_mx[1]].reshape(-1, 1)
    ##     xi = np.mgrid[xy_mn[0]:xy_mx[0]].reshape(1, -1)
    ##     pts = np.asarray((xi, yi)).T
    ##     contains = shape_obj.contains_pts(pts)
    ##     return contains

    def cutout_shape(self, shape_obj):
        view, mask = self.get_shape_view(shape_obj)

        data = self._slice(view)

        # mask non-containing members
        mdata = np.ma.array(data, mask=np.logical_not(mask))
        return mdata

    def get_shape_view(self, shape_obj, avoid_oob=True):
        """
        Calculate a bounding box in the data enclosing `shape_obj` and
        return a view that accesses it and a mask that is True only for
        pixels enclosed in the region.

        If `avoid_oob` is True (default) then the bounding box is clipped
        to avoid coordinates outside of the actual data.
        """
        x1, y1, x2, y2 = [int(np.round(n)) for n in shape_obj.get_llur()]

        if avoid_oob:
            # avoid out of bounds indexes
            xy_mn, xy_mx = self.viewer.get_limits()
            x1, x2 = max(xy_mn[0], x1), min(x2, xy_mx[0] - 1)
            y1, y2 = max(xy_mn[1], y1), min(y2, xy_mx[1] - 1)
            x1, y1, x2, y2 = [int(np.round(n)) for n in (x1, y1, x2, y2)]

        # calculate pixel containment mask in bbox
        xi, yi = np.meshgrid(range(x1, x2 + 1), range(y1, y2 + 1))
        pts = np.array((xi, yi)).T

        contains = shape_obj.contains_pts(pts)

        view = np.s_[y1:y2 + 1, x1:x2 + 1]
        return (view, contains)

    def get_pixels_on_line(self, x1, y1, x2, y2, getvalues=True):
        """Uses Bresenham's line algorithm to enumerate the pixels along
        a line.
        (see http://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm)

        If `getvalues`==False then it will return tuples of (x, y) coordinates
        instead of pixel values.
        """
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)

        dx = abs(x2 - x1)
        dy = abs(y2 - y1)
        if x1 < x2:
            sx = 1
        else:
            sx = -1
        if y1 < y2:
            sy = 1
        else:
            sy = -1
        err = dx - dy

        res = []
        x, y = x1, y1
        while True:
            res.append((x, y))
            if (x == x2) and (y == y2):
                break
            e2 = 2 * err
            if e2 > -dy:
                err = err - dy
                x += sx
            if e2 < dx:
                err = err + dx
                y += sy

        if getvalues:
            if np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2) > self.limit_cutout:
                return [self.getval_pt((x, y)) for x, y in res]

            x1, y1, x2, y2 = trcalc.sort_xy(x1, y1, x2, y2)
            data_np = self.cutout_data(x1, y1, x2 + 1, y2 + 1)
            res = [data_np[y - y1, x - x1] for x, y in res]

        return res

    def get_pixels_on_curve(self, curve_obj, getvalues=True):
        res = [(int(x), int(y))
               for x, y in curve_obj.get_points_on_curve(None)]

        if getvalues:
            ## x1, y1, x2, y2 = curve_obj.get_llur()
            ## x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
            ## data_np = self.cutout_data(x1, y1, x2 + 1, y2 + 1)
            ## return [data_np[y - y1, x - x1] for x, y in res]
            return [self.getval_pt((x, y)) for x, y in res]

        return res

    def _slice(self, view):
        # cutout our enclosing (possibly shortened) bbox
        x1, x2 = view[1].start, view[1].stop
        y1, y2 = view[0].start, view[0].stop
        xs = view[1].step
        if xs is None:
            xs = 1
        ys = view[0].step
        if ys is None:
            ys = 1

        data = self.cutout_data(x1, y1, x2, y2, xstep=xs, ystep=ys)
        return data

    @property
    def shape(self):
        wd, ht = self.get_size()[:2]
        return (ht, wd)

    @property
    def width(self):
        return self.get_size()[0]

    @property
    def height(self):
        return self.get_size()[1]

    @property
    def depth(self):
        return self.get_depth()

    @property
    def ndim(self):
        return len(self.shape)

    def get_size(self):
        xy_mn, xy_mx = self.viewer.get_limits()
        wd = int(abs(xy_mx[0] - xy_mn[0]))
        ht = int(abs(xy_mx[1] - xy_mn[1]))
        return (wd, ht)

    def get_depth(self):
        image = self.viewer.get_image()
        if image is not None:
            return image.get_depth()
        shape = self.shape
        if len(shape) > 2:
            return shape[-1]
        return 1

    def get_minmax(self, noinf=False):
        canvas = self.viewer.get_canvas()
        cvs_imgs = list(self.get_images([], canvas))
        if len(cvs_imgs) == 0:
            return (0, 0)
        mn, mx = cvs_imgs[0].get_image().get_minmax(noinf=noinf)
        for cvs_img in cvs_imgs[1:]:
            _mn, _mx = cvs_img.get_image().get_minmax(noinf=noinf)
            mn, mx = min(mn, _mn), max(mx, _mx)
        return (mn, mx)

    def get_shape(self):
        return self.shape

    def get_center(self):
        wd, ht = self.get_size()
        ctr_x, ctr_y = wd // 2, ht // 2
        return (ctr_x, ctr_y)

    def get_data_xy(self, x, y):
        x1, y1 = int(x), int(y)
        data = self.cutout_data(x1, y1, x1 + 1, y1 + 1)
        val = data[0, 0]
        if np.isnan(val):
            return None
        return val

    def info_xy(self, data_x, data_y, settings):

        image, pt = self.get_image_at_pt((data_x, data_y))
        ld_image = self.viewer.get_image()
        data_off = self.viewer.data_off

        if ld_image is not None:
            # if there is a loaded image, use it to establish info,
            # because the WCS will depend on that
            info = ld_image.info_xy(data_x, data_y, settings)
        else:
            info = Bunch.Bunch(itype='base', data_x=data_x, data_y=data_y,
                               x=data_x, y=data_y, value=None,
                               ra_txt="%+.3f" % (data_x),
                               dec_txt="%+.3f" % (data_y),
                               ra_lbl='X', dec_lbl='Y')

        if image is not None:
            info.image_x, info.image_y = pt
            _b_x, _b_y = pt[:2]
            _d_x, _d_y = (int(np.floor(_b_x + data_off)),
                          int(np.floor(_b_y + data_off)))
            info.value = image.get_data_xy(_d_x, _d_y)

        return info

    def pixtoradec(self, data_x, data_y, format='deg', coords='data'):
        ld_image = self.viewer.get_image()
        if ld_image is not None:
            # if there is a loaded image, then use it for WCS
            return ld_image.pixtoradec(data_x, data_y, format=format,
                                       coords=coords)

        # otherwise, look to see if there is an image under the data
        # point and use it's WCS if it has one
        image, pt = self.get_image_at_pt((data_x, data_y))
        if image is not None:
            return image.pixtoradec(pt[0], pt[1], format=format, coords=coords)

        raise ValueError("No image found for WCS conversion")

    def has_valid_wcs(self):
        ld_image = self.viewer.get_image()
        if ld_image is not None:
            return ld_image.has_valid_wcs()
        return False



#
# mode_base.py -- base class for defining modes
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import math

from ginga.events import ScrollEvent


class Mode:
    """Base class for defining a mode
    """

    # this should match that given in ~ginga.util.addons.show_mode_indicator()
    tag = '_$mode_indicator'

    @classmethod
    def is_compatible_viewer(cls, viewer):
        """Return True if this mode can be used with this viewer.
        """
        # raise NotImplementedError("Subclass should override this abstract class method!")
        return False

    def __init__(self, viewer, settings=None):
        super().__init__()

        self.viewer = viewer
        self.logger = viewer.get_logger()
        if settings is None:
            #settings = Settings.SettingGroup()
            settings = viewer.get_settings()
        self.settings = settings
        self.actions = dict()

        self._start_x = None
        self._start_y = None

    def get_docstring(self):
        import inspect

        # Insert section title at the beginning
        mode_name = self.__class__.__name__
        mode_mod = inspect.getmodule(self)
        _doc = "(No docstring found)"
        if self.__class__.__doc__ is not None:
            # prefer class docstring if found
            _doc = self.__class__.__doc__
        elif mode_mod.__doc__ is not None:
            # other
            _doc = mode_mod.__doc__
        mode_doc = ('{}\n{}\n'.format(mode_name, '=' * len(mode_name)) + _doc)
        return mode_name, mode_doc

    def get_settings(self):
        if self.settings is not None:
            return self.settings
        return self.viewer.get_settings()

    def start(self):
        """This method is called when the mode is explicitly invoked
        by the user.
        """
        raise NotImplementedError("subclass should implement this method")

    def stop(self):
        """This method is called when the mode is explicitly changed
        to a different mode (after having been started) by the user.
        """
        raise NotImplementedError("subclass should implement this method")

    def onscreen_message(self, msg, delay=None):
        self.viewer.onscreen_message(msg, delay=delay)

    def get_mode_line(self):
        canvas = self.viewer.get_private_canvas()
        indic = canvas.get_object_by_tag(self.tag)
        return indic

    def get_win_xy(self, viewer):
        x, y = viewer.get_last_win_xy()

        if not viewer.window_has_origin_upper():
            wd, ht = viewer.get_window_size()
            y = ht - y

        return x, y

    def get_direction(self, direction, rev=False):
        """
        Translate a direction in compass degrees into 'up' or 'down'.
        """
        # TODO: merge with the one in Bindings
        if (direction < 90.0) or (direction >= 270.0):
            if not rev:
                return 'up'
            else:
                return 'down'
        elif (90.0 <= direction < 270.0):
            if not rev:
                return 'down'
            else:
                return 'up'
        else:
            return 'none'

    def _pa_synth_scroll_event(self, event):

        dx, dy = float(event.delta_x), float(event.delta_y)
        amount = math.sqrt(dx ** 2.0 + dy ** 2.0)
        if dx == 0.0:
            if dy > 0:
                direction = 0.0
            else:
                direction = 180.0
        else:
            direction = math.atan(dy / dx)

        self.logger.debug("scroll amount=%f direction=%f" % (
            amount, direction))

        # synthesize a scroll event
        event = ScrollEvent(button=event.button, state=event.state,
                            mode=event.mode, modifiers=event.modifiers,
                            direction=direction, amount=amount,
                            data_x=event.data_x, data_y=event.data_y,
                            viewer=event.viewer)
        return event



#
# catalog.py -- image and star catalog interfaces for Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path
import tempfile
import re
import time
import warnings
from urllib.request import Request, urlopen, urlretrieve
from urllib.error import URLError, HTTPError

from ginga.misc import Bunch
from ginga.util import wcs

from astropy import coordinates, units

# Do we have astroquery >=0.3.5 installed?
have_astroquery = False
try:
    from astroquery.vo_conesearch import conesearch
    from astroquery.vizier import Vizier          # added astroquery.vizier to use vizier catalogs

    have_astroquery = True
    # limit the number of rows in the vizier query to 5000
    Vizier.ROW_LIMIT = 5000
except ImportError:
    pass

# these are modifed at bottom of this module
default_image_sources = []
default_catalog_sources = []
default_name_sources = []


class SourceError(Exception):
    """For exceptions raised by the `~ginga.util.catalog` module."""
    pass


class Star(object):
    def __init__(self, **kwdargs):
        starInfo = {}
        starInfo.update(kwdargs)
        self.starInfo = starInfo

    def __getitem__(self, key):
        return self.starInfo[key]

    def __contains__(self, key):
        return key in self.starInfo

    def __setitem__(self, key, value):
        self.starInfo[key] = value

    # TODO: Should deprecate this and encourage __contains__ like Python dict
    def has_key(self, key):
        return key in self.starInfo


class AstroqueryCatalogServer(object):
    """For queries using the ``astroquery.catalog`` function."""

    kind = 'astroquery.catalog'

    @classmethod
    def get_params_metadata(cls):
        from ginga.misc.ParamSet import Param
        return [
            Param(name='ra', type=str, widget='entry',
                  description="Right ascension component of center"),
            Param(name='dec', type=str, widget='entry',
                  description="Declination component of center"),
            Param(name='r', type=float, default=5.0, widget='entry',
                  description="Radius from center in arcmin"),
        ]

    def __init__(self, logger, full_name, key, querymod, mapping,
                 description=None, cat_columns=None, cat_column_filters=None):
        super(AstroqueryCatalogServer, self).__init__()
        if not have_astroquery:
            raise ImportError("'astroquery' not found, please install it")

        self.logger = logger
        self.full_name = full_name
        self.short_name = key
        self.mapping = mapping
        self.querymod = querymod
        self.cat_columns = cat_columns                 # added for usage with the vizier catalog
        self.cat_column_filters = cat_column_filters   # added for usage with the vizier catalog

        if description is None:
            description = full_name
        self.description = description

    def getParams(self):
        return self.get_params_metadata()

    def toStar(self, data, ext, magfield):
        try:
            mag = float(data[magfield])
        except Exception:
            mag = 0.0

        # Make sure we have at least these Ginga standard fields defined
        d = {'name': data[ext['id']],
             'ra_deg': float(data[ext['ra']]),
             'dec_deg': float(data[ext['dec']]),
             'mag': mag,
             'preference': 0.0,
             'priority': 0,
             'description': 'fake magnitude'}
        data.update(d)
        data['ra'] = wcs.ra_deg_to_str(data['ra_deg'])
        data['dec'] = wcs.dec_deg_to_str(data['dec_deg'])
        return Star(**data)

    def _search(self, center, radius, catalog, **kwargs):
        results = self.querymod.query_region(center, radius,
                                             catalog=catalog,
                                             **kwargs)
        if results is None:
            return results
        return results[0]

    def search(self, **params):
        """
        For compatibility with generic star catalog search.
        """

        self.logger.debug("search params=%s" % (str(params)))
        ra, dec = params['ra'], params['dec']
        if not (':' in ra):
            # Assume RA and DEC are in degrees
            ra_deg = float(ra)
            dec_deg = float(dec)
        else:
            # Assume RA and DEC are in standard string notation
            ra_deg = wcs.hmsStrToDeg(ra)
            dec_deg = wcs.dmsStrToDeg(dec)

        # Convert to degrees for search radius
        radius_deg = float(params['r']) / 60.0
        # radius_deg = float(params['r'])

        # Note requires astropy 0.3.x+
        c = coordinates.SkyCoord(ra_deg * units.degree,
                                 dec_deg * units.degree,
                                 frame='icrs')
        self.logger.info("Querying catalog: %s" % (self.full_name))
        time_start = time.time()
        with warnings.catch_warnings():  # Ignore VO warnings
            warnings.simplefilter('ignore')
            results = self._search(c, radius_deg * units.degree,
                                   self.full_name)

        time_elapsed = time.time() - time_start
        if results is None:
            self.logger.info("Null result in %.2f sec" % (
                time_elapsed))
            raise SourceError("Null result from query")
        else:
            numsources = len(results)
            self.logger.info("Found %d sources in %.2f sec" % (
                numsources, time_elapsed))

        # Scan the returned fields to find ones we need to extract
        # particulars from (ra, dec, id, magnitude)
        mags = []
        ext = {}
        fields = results.colnames
        for name in fields:
            if name == self.mapping['id']:
                ext['id'] = name
            elif name == self.mapping['ra']:
                ext['ra'] = name
            elif name == self.mapping['dec']:
                ext['dec'] = name
            if name in self.mapping.get('mag', []):
                mags.append(name)
        self.logger.debug("possible magnitude fields: %s" % str(mags))
        if len(mags) > 0:
            magfield = mags[0]
        else:
            magfield = None

        # prepare the result list
        starlist = []
        for i in range(numsources):
            source = dict(zip(fields, results[i]))
            starlist.append(self.toStar(source, ext, magfield))

        # metadata about the list
        columns = [('Name', 'name'),
                   ('RA', 'ra'),
                   ('DEC', 'dec'),
                   ('Mag', 'mag'),
                   ('Preference', 'preference'),
                   ('Priority', 'priority'),
                   ('Description', 'description'),
                   ]
        # Append extra columns returned by search to table header
        cols = list(fields)
        cols.remove(ext['ra'])
        cols.remove(ext['dec'])
        cols.remove(ext['id'])
        columns.extend(zip(cols, cols))

        # which column is the likely one to color source circles
        colorCode = 'Mag'

        info = Bunch.Bunch(columns=columns, color=colorCode)
        return starlist, info


class AstroqueryVOCatalogServer(AstroqueryCatalogServer):
    """For queries using the `astroquery.vo.conesearch` function."""

    kind = 'astroquery.vo_conesearch'

    def __init__(self, logger, full_name, key, mapping, description=None):
        super(AstroqueryVOCatalogServer, self).__init__(logger, full_name,
                                                        key, None, mapping,
                                                        description=description)

    def _search(self, center, radius, catalog):
        # override this methid to pass some special kwargs to the search
        results = conesearch.conesearch(center, radius, catalog_db=catalog,
                                        verbose=False,
                                        return_astropy_table=True,
                                        use_names_over_ids=False)
        return results


# add vizier catalog class
class AstroqueryVizierCatalogServer(AstroqueryCatalogServer):
    """For queries using the `astroquery.vizier` function."""

    kind = 'astroquery.vizier'

    @classmethod
    def get_params_metadata(cls):
        from ginga.misc.ParamSet import Param
        return [
            Param(name='ra', type=str, widget='entry',
                  description="Right ascension component of center"),
            Param(name='dec', type=str, widget='entry',
                  description="Declination component of center"),
            Param(name='width', type=float, default=1, widget='entry',
                  description="Width of box in degrees"),
            Param(name='height', type=float, default=1, widget='entry',
                  description="Height of box in degrees"),
            Param(name='r', type=float, default=5.0, widget='entry',
                  description="Radius from center in arcmin"),
        ]

    # add output colums and filters for vizier catalog
    def __init__(self, logger, full_name, key, mapping, cat_columns=[], cat_column_filters={}, description=None):
        super(AstroqueryVizierCatalogServer, self).__init__(logger, full_name,
                                                            key, None, mapping,
                                                            description=description, cat_columns=cat_columns, cat_column_filters=cat_column_filters)

    def set_cat_column_filters(self, cat_filter):
        self.cat_column_filters = cat_filter

    def set_cat_columns(self, cat_columns):
        self.cat_columns = cat_columns

    def getParams(self):
        return self.get_params_metadata()

    def _search_radius(self, center, radius, catalog, columns, column_filters):
        # override this method to pass some special kwargs to the search

        if columns:
            columns.insert(0, self.mapping['id'])
            columns.insert(1, self.mapping['ra'])
            columns.insert(2, self.mapping['dec'])

        v = Vizier(columns=columns, column_filters=column_filters)
        results = Vizier.query_region(center, radius, catalog=catalog)

        if len(results) > 0:
            return results[0]
        return None

    def _search_box(self, center, width, height, catalog, columns, column_filters):
        # override this method to pass some special kwargs to the search
        if columns:
            columns.insert(0, self.mapping['id'])
            columns.insert(1, self.mapping['ra'])
            columns.insert(2, self.mapping['dec'])

        v = Vizier(columns=columns, column_filters=column_filters)
        v.ROW_LIMIT = 500
        results = v.query_region(center, width=width, height=height, catalog=catalog)

        if len(results) > 0:
            return results[0]
        return None

    def search(self, **params):
        """
        For compatibility with generic star catalog search.
        """

        self.logger.debug("search params=%s" % (str(params)))
        ra, dec = params['ra'], params['dec']
        if not (':' in ra):
            # Assume RA and DEC are in degrees
            ra_deg = float(ra)
            dec_deg = float(dec)
        else:
            # Assume RA and DEC are in standard string notation
            ra_deg = wcs.hmsStrToDeg(ra)
            dec_deg = wcs.dmsStrToDeg(dec)

        # Note requires astropy 0.3.x+
        c = coordinates.SkyCoord(ra_deg * units.degree,
                                 dec_deg * units.degree,
                                 frame='icrs')

        self.logger.info("Querying catalog: %s" % (self.full_name))
        time_start = time.time()
        with warnings.catch_warnings():  # Ignore VO warnings
            warnings.simplefilter('ignore')
            try:
                if params['radius'] == 1:
                    # Convert to degrees for search radius
                    radius_deg = float(params['r']) / 60.0
                    # radius_deg = float(params['r'])
                    results = self._search_radius(c, radius_deg * units.degree,
                                                  self.full_name, columns=self.cat_columns, column_filters=self.cat_column_filters)
                elif params['box'] == 1:
                    # Convert to degrees for search width
                    width_deg = float(params['width']) / 60.0
                    # Convert to degrees for search height
                    height_deg = float(params['height']) / 60.0
                    results = self._search_box(c, width_deg * units.degree, height_deg * units.degree,
                                               self.full_name, columns=self.cat_columns, column_filters=self.cat_column_filters)
                else:
                    #attempt to use the radius, this shouldn't happen
                    # Convert to degrees for search radius
                    radius_deg = float(params['r']) / 60.0
                    # radius_deg = float(params['r'])
                    results = self._search_radius(c, radius_deg * units.degree,
                                                  self.full_name)
            except Exception as e:
                self.logger.error(f"vizier query raised an exception: {e}", exc_info=True)
                results = None

        time_elapsed = time.time() - time_start
        if results is None:
            self.logger.info("Null result in %.2f sec" % (
                time_elapsed))
            raise SourceError("Null result from query")
        else:
            numsources = len(results)
            self.logger.info("Found %d sources in %.2f sec" % (
                numsources, time_elapsed))

        # Scan the returned fields to find ones we need to extract
        # particulars from (ra, dec, id, magnitude)
        mags = []
        ext = {}        # {'id': 'recno', 'ra': 'RAJ2000', 'dec': 'DEJ2000'}
        fields = results.colnames
        for name in fields:
            if name == self.mapping['id']:
                ext['id'] = name
            elif name == self.mapping['ra']:
                ext['ra'] = name
            elif name == self.mapping['dec']:
                ext['dec'] = name
            if name in self.mapping.get('mag', []):
                mags.append(name)

        self.logger.debug("possible magnitude fields: %s" % str(mags))
        if len(mags) > 0:
            magfield = mags[0]
        else:
            magfield = None

        # prepare the result list
        starlist = []
        for i in range(numsources):
            source = dict(zip(fields, results[i]))
            starlist.append(self.toStar(source, ext, magfield))

        # metadata about the list
        columns = [('Name', 'name'),
                   ('RA', 'ra'),
                   ('DEC', 'dec'),
                   ]
        # Append extra columns returned by search to table header
        cols = list(fields)
        cols.remove(ext['id'])
        columns.extend(zip(cols, cols))

        # which column is the likely one to color source circles
        colorCode = 'Mag'

        info = Bunch.Bunch(columns=columns, color=colorCode)
        return starlist, info


class AstroqueryImageServer(object):
    """For queries using the ``astroquery.vo_conesearch`` function."""

    kind = 'astroquery.image'

    @classmethod
    def get_params_metadata(cls):
        from ginga.misc.ParamSet import Param
        return [
            Param(name='ra', type=str, default='', widget='entry',
                  description="Right ascension component of center"),
            Param(name='dec', type=str, widget='entry',
                  description="Declination component of center"),
            Param(name='width', type=float, default=1, widget='entry',
                  description="Width of box in degrees"),
            Param(name='height', type=float, default=1, widget='entry',
                  description="Height of box in degrees"),
        ]

    def __init__(self, logger, full_name, key, querymod, description=None):
        super(AstroqueryImageServer, self).__init__()
        if not have_astroquery:
            raise ImportError('astroquery not found, please install astroquery')

        self.logger = logger
        self.full_name = full_name
        self.short_name = key
        if isinstance(querymod, str) and querymod.lower() == 'skyview':
            from astroquery.skyview import SkyView
            self.querymod = SkyView
        else:
            self.querymod = querymod
        if description is None:
            description = full_name
        self.description = description

    def getParams(self):
        return self.get_params_metadata()

    def _search(self, center, wd_deg, ht_deg):
        survey = [self.short_name]

        results = self.querymod.get_image_list(center, survey,
                                               width=wd_deg * units.degree,
                                               height=ht_deg * units.degree)
        return results

    # TODO: dstpath provides the pathname for storing the image
    def search(self, dstpath, **params):
        """
        For compatibility with generic image catalog search.
        """

        self.logger.debug("search params=%s" % (str(params)))
        ra, dec = params['ra'], params['dec']
        if not (':' in ra):
            # Assume RA and DEC are in degrees
            ra_deg = float(ra)
            dec_deg = float(dec)
        else:
            # Assume RA and DEC are in standard string notation
            ra_deg = wcs.hmsStrToDeg(ra)
            dec_deg = wcs.dmsStrToDeg(dec)

        # Convert to degrees for search
        wd_deg = float(params['width']) / 60.0
        ht_deg = float(params['height']) / 60.0

        # Note requires astropy 3.x+
        c = coordinates.SkyCoord(ra_deg * units.degree,
                                 dec_deg * units.degree,
                                 frame='icrs')
        self.logger.info("Querying image source: %s" % (self.full_name))
        time_start = time.time()
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            results = self._search(c, wd_deg, ht_deg)

        time_elapsed = time.time() - time_start
        if results is None:
            self.logger.info("Null result in %.2f sec" % (
                time_elapsed))
            return None
        else:
            numsources = len(results)
            self.logger.info("Found %d images in %.2f sec" % (
                numsources, time_elapsed))
            if numsources == 0:
                self.logger.warning("Found no images in this area" % len(results))
                return None

        # For now, we pick the first one found
        urls = list(results)
        # fitspath = results[0].make_dataset_filename(dir=tempfile.gettempdir())

        # file will be downloaded
        fitspath = urls[0]

        # explicit return
        return fitspath


class AstroqueryNameServer(object):
    """For object name lookups using `astroquery`"""

    kind = 'astroquery.names'

    def __init__(self, logger, full_name, key, mapping, description=None):
        super(AstroqueryNameServer, self).__init__()
        if not have_astroquery:
            raise ImportError("'astroquery' not found, please install it")

        self.logger = logger
        self.full_name = full_name
        self.short_name = key
        self.mapping = mapping
        if description is None:
            description = full_name
        self.description = description

    def search(self, name, **kwargs):
        if self.short_name == 'SIMBAD':
            from astroquery.simbad import Simbad
            results = Simbad.query_object(name, **kwargs)
            if results is None:
                raise SourceError("No results found for name '{}'".format(name))

            # from SIMBAD, coords come formatted as a string
            ra = ':'.join(results['RA'][0].split())
            dec = ':'.join(results['DEC'][0].split())

        elif self.short_name == 'NED':
            from astroquery.ned import Ned
            results = Ned.query_object(name, **kwargs)
            if results is None:
                # Ned usually returns an exception for non-found objects
                # but lets put this test just in case
                raise SourceError("No results found for name '{}'".format(name))

            # from NED, coords come as degrees in float
            ra = wcs.ra_deg_to_str(results['RA'][0])
            dec = wcs.dec_deg_to_str(results['DEC'][0])

        else:
            raise SourceError("Don't know how to query source '{}'".format(self.short_name))

        return ra, dec


class URLServer(object):

    kind = 'url'

    def __init__(self, logger, full_name, key, url, description):
        self.logger = logger
        self.full_name = full_name
        self.short_name = key
        self.base_url = url
        self.reqtype = 'get'
        self.description = description

        self.params = self._parse_params(url)

    def _parse_params(self, url):
        params = {}
        regex = r'^.*?\%\((\w+)\)([sfd])(.*)$'
        match = re.match(regex, url)
        count = 0
        while match:
            key, typ, sfx = match.groups()
            idx = ['s', 'd', 'f'].index(typ)
            cvt = (str, int, float)[idx]
            params[key] = Bunch.Bunch(name=key, convert=cvt, order=count)
            count += 1
            match = re.match(regex, sfx)

        return params

    def getParams(self):
        return self.params

    def convertParams(self, params):
        d = {}
        for key, bnch in self.params.items():
            d[key] = bnch.convert(params[key])
        return d

    def fetch(self, url, filepath=None):
        data = ""

        req = Request(url)

        try:
            self.logger.info("Opening url=%s" % (url))
            try:
                response = urlopen(req)  # nosec

            except HTTPError as e:
                self.logger.error("Server returned error code %s" % (e.code))
                raise e
            except URLError as e:
                self.logger.error("Server URL failure: %s" % (str(e.reason)))
                raise e
            except Exception as e:
                self.logger.error("URL fetch failure: %s" % (str(e)))
                raise e

            self.logger.debug("getting HTTP headers")
            info = response.info()  # noqa

            self.logger.debug("getting data")
            data = response.read()
            self.logger.debug("fetched %d bytes" % (len(data)))
            # data = data.decode('ascii')

        except Exception as e:
            self.logger.error("Error reading data from '%s': %s" % (
                url, str(e)))
            raise e

        if filepath:
            with open(filepath, 'wb') as out_f:
                out_f.write(data)
            return None

        else:
            return data

    def retrieve(self, url, filepath=None, cb_fn=None):
        ofilepath = filepath
        if (filepath is None) or os.path.isdir(filepath):
            with tempfile.NamedTemporaryFile(dir=filepath,
                                             delete=False) as out_f:
                filepath = out_f.name

        try:
            self.logger.info("Opening url=%s" % (url))

            if cb_fn is not None:
                localpath, info = urlretrieve(url, filepath, cb_fn)  # nosec
            else:
                localpath, info = urlretrieve(url, filepath)  # nosec

        except Exception as e:
            self.logger.error("URL fetch failure: %s" % (str(e)))
            raise e

        if ofilepath is not None:
            return localpath

        with open(filepath, 'r') as in_f:
            return in_f.read()

    def search(self, filepath, **params):
        url = self.base_url % params

        self.fetch(url, filepath=filepath)
        return filepath


class ImageServer(URLServer):

    kind = 'ginga.image'

    @classmethod
    def get_params_metadata(cls):
        from ginga.misc.ParamSet import Param
        return [
            Param(name='ra', type=str, default='', widget='entry',
                  description="Right ascension component of center"),
            Param(name='dec', type=str, widget='entry',
                  description="Declination component of center"),
            Param(name='width', type=float, default=1, widget='entry',
                  description="Width of box in degrees"),
            Param(name='height', type=float, default=1, widget='entry',
                  description="Height of box in degrees"),
        ]

    def __init__(self, logger, full_name, key, url, description):
        super(ImageServer, self).__init__(logger, full_name, key, url,
                                          description)


class CatalogServer(URLServer):

    kind = 'ginga.catalog'

    @classmethod
    def get_params_metadata(cls):
        from ginga.misc.ParamSet import Param
        return [
            Param(name='ra', type=str, widget='entry',
                  description="Right ascension component of center"),
            Param(name='dec', type=str, widget='entry',
                  description="Declination component of center"),
            Param(name='r', type=float, default=1.0, widget='entry',
                  description="Radius from center in arcmin"),
            Param(name='r2', type=float, default=2.0, widget='entry',
                  description="Outer radius from center in arcmin"),
        ]

    def __init__(self, logger, full_name, key, url, description):
        super(CatalogServer, self).__init__(logger, full_name, key, url,
                                            description)
        self.index = {'name': 0, 'ra': 1, 'dec': 2, 'mag': 10}
        self.format = 'str'
        self.equinox = 2000.0

    def set_index(self, **kwdargs):
        self.index.update(kwdargs)

    def search(self, **params):
        self.logger.debug("search params=%s" % (str(params)))
        url = self.base_url % params

        data = self.fetch(url, filepath=None)
        data = data.decode("utf8")

        lines = data.split('\n')
        offset = 0
        while offset < len(lines):
            line = lines[offset].strip()
            offset += 1
            if line.startswith('-'):
                break
        self.logger.debug("offset=%d" % (offset))

        results = []
        table = [lines[offset - 2]]

        for line in lines[offset:]:
            line = line.strip()
            if (len(line) == 0) or line.startswith('#'):
                continue
            elts = line.split()
            if (len(elts) < 3):
                continue
            table.append(line)

            try:
                name = elts[self.index['name']]
                ra = elts[self.index['ra']]
                dec = elts[self.index['dec']]
                mag = float(elts[self.index['mag']])

                if (self.format == 'deg') or not (':' in ra):
                    # Assume RA and DEC are in degrees
                    ra_deg = float(ra)
                    dec_deg = float(dec)
                else:
                    # Assume RA and DEC are in standard string notation
                    ra_deg = wcs.hmsStrToDeg(ra)
                    dec_deg = wcs.dmsStrToDeg(dec)

                # convert ra/dec via EQUINOX change if catalog EQUINOX is
                # not the same as our default one (2000)
                if self.equinox != 2000.0:
                    ra_deg, dec_deg = wcs.eqToEq2000(ra_deg, dec_deg,
                                                     self.equinox)

                ra_txt = wcs.ra_deg_to_str(ra_deg)
                dec_txt = wcs.dec_deg_to_str(dec_deg)
                self.logger.debug("STAR %s AT ra=%s dec=%s mag=%f" % (
                    name, ra_txt, dec_txt, mag))

                results.append(Star(name=name, ra_deg=ra_deg, dec_deg=dec_deg,
                                    ra=ra_txt, dec=dec_txt, mag=mag,
                                    preference=0.0, priority=0,
                                    description=''))

            except Exception as e:
                self.logger.error("Error parsing catalog query results: %s" % (
                    str(e)))

        # metadata about the list
        columns = [('Name', 'name'),
                   ('RA', 'ra'),
                   ('DEC', 'dec'),
                   ('Mag', 'mag'),
                   ('Preference', 'preference'),
                   ('Priority', 'priority'),
                   ('Description', 'description'),
                   ]
        info = Bunch.Bunch(columns=columns, color='Mag')

        return (results, info)


class ServerBank(object):

    def __init__(self, logger):
        self.logger = logger

        self.clear()

    def clear(self):
        self.imbank = {}
        self.ctbank = {}
        self.nmbank = {}

    def add_image_server(self, srvobj):
        self.imbank[srvobj.short_name] = srvobj

    def add_catalog_server(self, srvobj):
        self.ctbank[srvobj.short_name] = srvobj

    def add_name_server(self, srvobj):
        self.nmbank[srvobj.short_name] = srvobj

    def get_image_server(self, key):
        return self.imbank[key]

    def get_catalog_server(self, key):
        return self.ctbank[key]

    def get_name_server(self, key):
        return self.nmbank[key]

    def get_server_names(self, kind='image'):
        if kind == 'image':
            keys = self.imbank.keys()
        elif kind == 'name':
            keys = self.nmbank.keys()
        else:
            keys = self.ctbank.keys()
        keys = list(keys)
        keys.sort()
        return keys

    def get_image(self, key, filepath, **params):
        obj = self.imbank[key]

        return obj.search(filepath, **params)

    def get_catalog(self, key, filepath, **params):
        obj = self.ctbank[key]

        return obj.search(**params)


# ---- SET UP DEFAULT SOURCES ----

if have_astroquery:
    # set up default name sources, catalog sources and image sources

    default_name_sources.extend([
        {'shortname': "SIMBAD", 'fullname': "SIMBAD",
         'type': 'astroquery.names'},
        {'shortname': "NED", 'fullname': "NED",
         'type': 'astroquery.names'},
    ])

    default_catalog_sources.extend([
        {'shortname': "GSC 2.3",
         'fullname': "Guide Star Catalog 2.3 Cone Search 1",
         'type': 'astroquery.vo_conesearch',
         'mapping': {'id': 'objID', 'ra': 'ra', 'dec': 'dec', 'mag': ['Mag']}},
        {'shortname': "USNO-A2.0 1",
         'fullname': "The USNO-A2.0 Catalogue 1",
         'type': 'astroquery.vo_conesearch',
         'mapping': {'id': 'USNO-A2.0', 'ra': 'RAJ2000', 'dec': 'DEJ2000',
                     'mag': ['Bmag', 'Rmag']}},
        {'shortname': "2MASS 1",
         'fullname': "Two Micron All Sky Survey (2MASS) 1",
         'type': 'astroquery.vo_conesearch',
         'mapping': {'id': 'htmID', 'ra': 'ra', 'dec': 'dec',
                     'mag': ['h_m', 'j_m', 'k_m']}},
        {'shortname': "APASS DR9",
         'fullname': "II/336/apass9",
         'type': 'astroquery.vizier',
         'mapping': {'id': 'recno', 'ra': 'RAJ2000', 'dec': 'DEJ2000', 'mag': ['Vmag', 'B-V', 'Bmag', 'g\'mag', 'r\'mag', 'i\'mag']}},
    ])

    default_image_sources.extend([
        {'shortname': "DSS",
         'fullname': "Digital Sky Survey 1",
         'type': 'astroquery.image',
         'source': 'skyview'},
        {'shortname': "DSS1 Blue",
         'fullname': "Digital Sky Survey 1 Blue",
         'type': 'astroquery.image',
         'source': 'skyview'},
        {'shortname': "DSS1 Red",
         'fullname': "Digital Sky Survey 1 Red",
         'type': 'astroquery.image',
         'source': 'skyview'},
        {'shortname': "DSS2 Red",
         'fullname': "Digital Sky Survey 2 Red",
         'type': 'astroquery.image',
         'source': 'skyview'},
        {'shortname': "DSS2 Blue",
         'fullname': "Digital Sky Survey 2 Blue",
         'type': 'astroquery.image',
         'source': 'skyview'},
        {'shortname': "DSS2 IR",
         'fullname': "Digital Sky Survey 2 Infrared",
         'type': 'astroquery.image',
         'source': 'skyview'},
    ])



#
# PlotView.py -- base class for plot viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time
import threading
from io import BytesIO

import numpy as np

from ginga import Mixins, colors
from ginga.util.viewer import ViewerBase
from ginga.misc import Bunch, Settings
from ginga.plot.Plotable import Plotable
from ginga.canvas.CanvasObject import get_canvas_types
from ginga import Bindings
from ginga.fonts import font_asst

try:
    from matplotlib.figure import Figure

    from ginga.gw.Plot import PlotWidget, PlotEventMixin
    have_mpl = True
except ImportError:
    have_mpl = False

if not have_mpl:
    # NOTE: needed for Sphinx documentation building
    class PlotEventMixin:  # noqa
        pass

__all__ = ['PlotViewBase', 'CanvasView', 'PlotViewGw']


class PlotViewBase(ViewerBase):
    """A Ginga viewer for displaying 2D plots using matplotlib.
    """

    vname = 'Ginga Plot'
    vtypes = [Plotable]

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer."""
        if have_mpl and isinstance(dataobj, Plotable):
            return True
        return False

    def __init__(self, logger=None, settings=None, figure=None):

        if not have_mpl:
            raise ImportError("Install 'matplotlib' to use this viewer")

        ViewerBase.__init__(self, logger=logger, settings=settings)
        self.needs_scrolledview = True

        self.t_ = self.settings
        # misc
        self.t_.add_defaults(resize_lagtime=0.25)

        self.t_.add_defaults(plot_bg='white',
                             plot_show_mode=False,
                             plot_show_grid=True,
                             plot_show_legend=False,
                             plot_save_suffix='.png',
                             plot_dpi=100, plot_save_dpi=100,
                             plot_title_fontsize=None,
                             plot_axis_fontsize=None,
                             plot_limits=None,
                             plot_range=None,
                             plot_enter_focus=True,
                             plot_autozoom='override',
                             plot_zoom_rate=1.1)

        self.t_.get_setting('plot_range').add_callback('set', self.ranges_change_cb)
        self.t_.get_setting('plot_show_grid').add_callback('set', self.set_grid_cb)
        self.t_.get_setting('plot_show_legend').add_callback('set', self.set_legend_cb)

        # pan position
        self.t_.add_defaults(plot_pan=(0.0, 0.0))
        self.t_.get_setting('plot_pan').add_callback('set', self.pan_change_cb)

        # for axis scaling
        self.t_.add_defaults(plot_dist_axis=('linear', 'linear'))
        for name in ['plot_dist_axis']:
            self.t_.get_setting(name).add_callback('set', self.axis_dist_change_cb)

        # plot markers
        self.t_.add_defaults(plot_show_marker=False, plot_marker_size=5,
                             plot_marker_width=0.35,
                             plot_marker_color='blue',
                             plot_marker_style='o')
        for name in ['plot_show_marker', 'plot_marker_size',
                     'plot_marker_width', 'plot_marker_color',
                     'plot_marker_style']:
            self.t_.get_setting(name).add_callback('set',
                                                   lambda setting, value: self.replot())

        # embedded profiles
        self.use_embedded_profile = True
        self.t_.add_defaults(profile_use_range=True)
        self.profile_keylist = ['plot_range']
        for name in self.profile_keylist:
            self.t_.get_setting(name).add_callback('set',
                                                   self._update_profile_cb)

        # viewer profile support
        self.default_viewer_profile = None
        self.t_.add_defaults(viewer_restore_range=True,
                             viewer_restore_pan=False)
        self.capture_default_viewer_profile()

        # optimization of resizing
        self.time_last_resize = time.time()
        self.resize_lagtime = self.t_.get('resize_lagtime', 0.25)
        self._resize_lock = threading.RLock()
        self._resize_dims = (0, 0)

        self._dataobj = None
        self._data_limits = None
        self.rgb_order = 'RGBA'
        self.last_data_x, self.last_data_y = 0, 0
        # cursors
        self.cursor = {}

        # For callbacks
        for name in ['image-set', 'image-unset',
                     'limits-set', 'range-set', 'redraw',
                     'configure']:
            self.enable_callback(name)

        if figure is None:
            figure = Figure()
        self.figure = figure

        self.artist_dct = dict()
        bg = self.settings.get('plot_bg', 'white')
        self.ax = self.figure.add_subplot(111, facecolor=bg)
        # self.ax.grid(True)

        self.dc = get_canvas_types()
        self.private_canvas = self.dc.DrawingCanvas()

        # setup default fg color
        color = self.t_.get('color_fg', "#D0F0E0")
        r, g, b = colors.resolve_color(color)
        self.clr_fg = (r, g, b)

        # setup default bg color
        color = self.t_.get('color_bg', "#000000")  # #404040
        r, g, b = colors.resolve_color(color)
        self.clr_bg = (r, g, b)

        # holds onscreen text object
        self._ost = None
        # these will get set in the event mixin
        self.timer_resize = None
        self.timer_msg = None

        self.add_callback('configure', self.resize_cb)
        self.plot_w = PlotWidget(self)

    def set_figure(self, figure, mpl_canvas=None):
        self.figure = figure
        if mpl_canvas is None:
            mpl_canvas = self.figure.canvas

        dpi = figure.get_dpi()
        if dpi is None or dpi < 0.1:
            dpi = self.t_['plot_dpi']
            wd_px, ht_px = self.plot_w.get_size()
            wd_in, ht_in = float(wd_px) / dpi, float(ht_px) / dpi
            figure.set_size_inches(wd_in, ht_in)

        if hasattr(self.figure, 'set_tight_layout'):
            self.figure.set_tight_layout(True)

    def get_figure(self):
        return self.figure

    def get_widget(self):
        return self.figure.canvas

    def get_ginga_widget(self):
        return self.plot_w

    def get_window_size(self):
        fig = self.get_figure()
        fig_wd_in, fig_ht_in = fig.get_size_inches()
        fig_dpi = fig.get_dpi()
        wd_px = int(fig_wd_in * fig_dpi)
        ht_px = int(fig_ht_in * fig_dpi)
        return (wd_px, ht_px)

    def add_axis(self, **kwdargs):
        if self.ax is not None:
            ax, self.ax = self.ax, None
            ax.remove()
        self.ax = self.figure.add_subplot(111, **kwdargs)
        return self.ax

    def get_axis(self):
        if self.ax is None:
            return self.add_axis()
        return self.ax

    def set_dataobj(self, dataobj):
        """Set 1D data to be displayed.

        If there is no error, the ``'image-unset'`` and ``'image-set'``
        callbacks will be invoked.

        Parameters
        ----------
        dataobj : `~ginga.AstroImage.AstroImage` or `~ginga.plot.Plotable.Plotable`
            AstroImage object or Plotable.

        """
        if not self.viewable(dataobj):
            raise ValueError("Wrong type of object to load: %s" % (
                str(type(dataobj))))

        old_dataobj = self._dataobj
        if old_dataobj is not None:
            self.make_callback('image-unset', old_dataobj)

        self.clear()

        self._dataobj = dataobj

        if isinstance(dataobj, Plotable):
            dataobj.add_callback('modified', self._dataobj_modified_cb)
            self.show_plotable(dataobj)

        self.make_callback('image-set', dataobj)

    def get_dataobj(self):
        return self._dataobj

    def get_data_size(self):
        # NOTE: used to determine widths needed in cursor readout
        (x_lo, y_lo), (x_hi, y_hi) = self.get_limits()
        return (x_hi, y_hi)

    def set_window_size(self, wd_px, ht_px):
        with self._resize_lock:
            self._resize_dims = (wd_px, ht_px)
            self.time_last_resize = time.time()
            self._resize_flag = False

            self.logger.debug("canvas resized to %dx%d" % (wd_px, ht_px))
            fig = self.get_figure()
            fig.set_size_inches(float(wd_px) / fig.dpi, float(ht_px) / fig.dpi)

        self.make_callback('configure', wd_px, ht_px)

    configure = set_window_size
    configure_window = set_window_size

    def resize_cb(self, viewer, wd_px, ht_px):
        self.logger.debug(f"viewer resized to {wd_px}x{ht_px}")
        self._set_variable_font_sizes()

        self.redraw()

    def delayed_resize(self):
        """Called when the timer for a delayed resize expires."""
        with self._resize_lock:
            if not self._resize_flag:
                # <-- no deferred resize found
                return
            wd_px, ht_px = self._resize_dims
            self.set_window_size(wd_px, ht_px)

    def reschedule_resize(self, wd_px, ht_px):
        with self._resize_lock:
            if time.time() - self.time_last_resize > self.resize_lagtime:
                self.set_window_size(wd_px, ht_px)
            else:
                if self.timer_resize is None:
                    self.set_window_size(wd_px, ht_px)
                else:
                    self._resize_flag = True
                    self._resize_dims = (wd_px, ht_px)
                    self.timer_resize.cond_set(self.resize_lagtime)

    def clear(self, redraw=True):
        """Clear plot display."""
        self.logger.debug('clearing viewer...')
        self._dataobj = None
        self.clear_data()
        if redraw:
            self.redraw_now()

    def clear_data(self):
        self.artist_dct = dict()
        self.ax.set_aspect('auto')
        #self.ax.set_aspect('equal', adjustable='box')
        self.ax.cla()
        self.t_.set(plot_limits=None, plot_range=None)
        self._data_limits = None

    def remove_artist_category(self, artist_category, redraw=True):
        artists = self.artist_dct.setdefault(artist_category, [])
        for artist in artists:
            try:
                artist.remove()
            except Exception as e:
                pass
        if redraw:
            self.redraw()

    def _set_variable_font_sizes(self):
        # recalculate font sizes if left to variable setting
        width, height = self.get_window_size()
        title_font_size = self.t_['plot_title_fontsize']
        if title_font_size is None:
            title_font_size = font_asst.calc_font_size(width)

        axis_font_size = self.t_['plot_axis_fontsize']
        if axis_font_size is None:
            axis_font_size = font_asst.calc_font_size(width)

        try:
            ax = self.ax
            ax.title.set_fontsize(title_font_size)
            for item in ([ax.xaxis.get_label(), ax.yaxis.get_label()] +
                         ax.get_xticklabels() + ax.get_yticklabels()):
                item.set(fontsize=axis_font_size)

            self.redraw()
        except Exception as e:
            self.logger.error(f"error changing font size: {e}", exc_info=True)

    def set_titles(self, title=None, x_axis=None, y_axis=None, redraw=True):
        # recalculate font sizes if left to variable setting
        width, height = self.get_window_size()
        title_font_size = self.t_['plot_title_fontsize']
        if title_font_size is None:
            title_font_size = font_asst.calc_font_size(width)

        axis_font_size = self.t_['plot_axis_fontsize']
        if axis_font_size is None:
            axis_font_size = font_asst.calc_font_size(width)

        plotable = self.get_dataobj()

        if x_axis is not None:
            if isinstance(plotable, Plotable):
                plotable.set(x_axis_title=x_axis)
            self.ax.set_xlabel(x_axis, fontsize=axis_font_size)
        if y_axis is not None:
            if isinstance(plotable, Plotable):
                plotable.set(y_axis_title=y_axis)
            self.ax.set_ylabel(y_axis, fontsize=axis_font_size)
        if title is not None:
            if isinstance(plotable, Plotable):
                plotable.set(title=title)
            self.ax.set_title(title, fontsize=title_font_size)

        # Make X axis labels a little more readable
        lbls = self.ax.xaxis.get_ticklabels()
        for lbl in lbls:
            lbl.set(rotation=45, horizontalalignment='right',
                    fontsize=axis_font_size)

        # update Y axis labels
        lbls = self.ax.yaxis.get_ticklabels()
        for lbl in lbls:
            lbl.set(fontsize=axis_font_size)

        if redraw:
            self.redraw()

    def set_grid(self, tf):
        self.t_.set(plot_show_grid=tf)

    def set_grid_cb(self, setting, tf):
        # record in plotable
        plotable = self.get_dataobj()
        if isinstance(plotable, Plotable):
            plotable.set(grid=tf)

        self.ax.grid(tf)
        self.redraw()

    def toggle_grid(self):
        tf = self.t_.get('plot_show_grid', False)
        self.set_grid(not tf)

    def set_legend(self, tf, redraw=True):
        self.t_.set(plot_show_legend=tf)

    def set_legend_cb(self, setting, tf):
        # record in plotable
        plotable = self.get_dataobj()
        if isinstance(plotable, Plotable):
            plotable.set(legend=tf)

        legend = self.ax.get_legend()
        if tf:
            if legend is None:
                self.ax.legend(loc='upper right')
        else:
            if legend is not None:
                legend.remove()
        self.redraw()

    def toggle_legend(self):
        tf = self.t_.get('plot_show_legend', False)
        self.set_legend(not tf)

    def _record_limits(self, x_data, y_data):
        x_min, x_max = np.nanmin(x_data), np.nanmax(x_data)
        y_min, y_max = np.nanmin(y_data), np.nanmax(y_data)

        adjusted = False
        if self._data_limits is None:
            x_lo, y_lo, x_hi, y_hi = x_min, y_min, x_max, y_max
            adjusted = True
        else:
            (x_lo, y_lo), (x_hi, y_hi) = self._data_limits
            if x_min < x_lo:
                x_lo, adjusted = x_min, True
            if x_max > x_hi:
                x_hi, adjusted = x_max, True
            if y_min < y_lo:
                y_lo, adjusted = y_min, True
            if y_max > y_hi:
                y_hi, adjusted = y_max, True

        if adjusted:
            self._data_limits = [(x_lo, y_lo), (x_hi, y_hi)]

    def _plot_line(self, x_data, y_data, artist_category='default', name=None,
                   linewidth=1, linestyle='-', color='black',
                   alpha=1.0):
        """Simple line plot."""

        plt_kw = {'lw': linewidth,
                  'ls': linestyle,
                  'color': color,
                  'alpha': alpha,
                  'antialiased': True,
                  'label': name,
                  }
        if self.t_['plot_show_marker']:
            plt_kw.update({'marker': self.t_['plot_marker_style'],
                           'ms': self.t_['plot_marker_size'],
                           'mew': self.t_['plot_marker_width'],
                           'mfc': color,
                           'mec': color})

        try:
            artists = self.artist_dct.setdefault(artist_category, [])
            line, = self.ax.plot(x_data, y_data, **plt_kw)
            artists.append(line)

            # adjust limits if necessary
            self._record_limits(x_data, y_data)

        except Exception as e:
            self.logger.error(str(e), exc_info=True)

    def _plot_text(self, x, y, text, artist_category='default', rot_deg=0,
                   linewidth=1, color='black', font='sans', fontsize=12,
                   bgcolor=None, bordercolor=None, borderlinewidth=0,
                   borderpadding=4):
        artists = self.artist_dct.setdefault(artist_category, [])
        fontdict = dict(color=color, family=font, size=fontsize)
        kwargs = dict()
        if bgcolor is not None:
            bbox = dict(facecolor=bgcolor)
            kwargs['bbox'] = bbox
        if bordercolor is not None:
            bbox = kwargs.get('bbox', dict())
            bbox['edgecolor'] = bordercolor
            bbox['linewidth'] = borderlinewidth
        bbox = kwargs.get('bbox', None)
        if bbox is not None:
            bbox['pad'] = borderpadding

        text = self.ax.text(x, y, text, color=color, rotation=rot_deg,
                            fontdict=fontdict, clip_on=True, **kwargs)
        artists.append(text)

        # adjust limits if necessary
        self._record_limits(x, y)

    def _plot_xrange(self, x_lo, x_hi, artist_category='default',
                     linewidth=0, fillcolor='aquamarine', fillalpha=0.5):
        artists = self.artist_dct.setdefault(artist_category, [])
        hex_color = colors.resolve_color(fillcolor, format='tuple')

        span = self.ax.axvspan(x_lo, x_hi, facecolor=hex_color, alpha=fillalpha)
        artists.append(span)

        # adjust limits if necessary
        # self._record_limits(x, y)

    def _plot_yrange(self, y_lo, y_hi, artist_category='default',
                     linewidth=0, fillcolor='aquamarine', fillalpha=0.5):
        artists = self.artist_dct.setdefault(artist_category, [])
        hex_color = colors.resolve_color(fillcolor, format='tuple')

        span = self.ax.ayvspan(y_lo, y_hi, facecolor=hex_color, alpha=fillalpha)
        artists.append(span)

        # adjust limits if necessary
        # self._record_limits(x, y)

    def set_limits(self, limits):
        self.settings.set(plot_limits=limits)

        # NOTE: make compatible callback to image viewer
        self.make_callback('limits-set', limits)

    def get_limits(self):
        limits = self.settings['plot_limits']

        if limits is None:
            # No user defined limits. Return limits set from data points.
            if self._data_limits is None:
                return [[0.0, 0.0], [0.0, 0.0]]

            limits = self._data_limits.copy()

        return limits

    def reset_limits(self):
        """Reset the bounding box of the viewer extents.

        Parameters
        ----------
        None
        """
        self.t_.set(plot_limits=None)

    def get_pan_bbox(self):
        """Get the coordinates in the actual data corresponding to the
        area shown in the display for the current zoom level and pan.

        Returns
        -------
        points : list
            Coordinates in the form of
            ``[(x0, y0), (x1, y1), (x2, y2), (x3, y3)]``
            from lower-left to lower-right.

        """
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        return [(x_lo, y_lo), (x_hi, y_lo), (x_hi, y_hi), (x_lo, y_hi)]

    def set_ranges(self, x_range=None, y_range=None):
        adjusted = False
        ranges = self.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        if x_range is not None:
            x_lo, x_hi = x_range
            ranges[0] = [x_lo, x_hi]
            adjusted = True

        if y_range is not None:
            y_lo, y_hi = y_range
            ranges[1] = [y_lo, y_hi]
            adjusted = True

        if adjusted:
            self.t_.set(plot_range=ranges, callback=True)

    def ranges_change_cb(self, setting, ranges):
        adjusted = False
        if ranges is None:
            ranges = (None, None)
        x_range, y_range = ranges

        ranges = self.get_ranges()
        (x_lo, x_hi), (y_lo, y_hi) = ranges

        if x_range is not None:
            x_lo, x_hi = x_range
            ranges[0] = [x_lo, x_hi]
            adjusted = True
            self.ax.set_xlim(x_lo, x_hi)

        if y_range is not None:
            y_lo, y_hi = y_range
            ranges[1] = [y_lo, y_hi]
            adjusted = True
            self.ax.set_ylim(y_lo, y_hi)

        if adjusted:
            self.make_callback('range-set', ranges.copy())

            _pan_x, _pan_y = self.t_['plot_pan'][:2]
            pan_x, pan_y = (x_lo + x_hi) * 0.5, (y_lo + y_hi) * 0.5
            if not np.isclose(_pan_x, pan_x) or not np.isclose(_pan_y, pan_y):
                # need to set the pan position because range is changing
                # not according to pan position
                #self.t_['plot_pan'] = (pan_x, pan_y)
                self.set_pan(pan_x, pan_y)

        self.redraw_now()

    def get_ranges(self):
        ranges = self.settings['plot_range']
        if ranges is None:
            (x_lo, y_lo), (x_hi, y_hi) = self.get_limits()
            return [[x_lo, x_hi], [y_lo, y_hi]]

        return ranges.copy()

    def zoom_fit(self, axis='xy', no_reset=False):
        """Zoom to fit display window.
        Pan the image and scale the view to fit the size of the set
        limits (usually set to the image size).  Parameter `axis` can
        be used to set which axes are allowed to be scaled;
        Also see :meth:`zoom_to`.

        Parameters
        ----------
        axis : str
            One of: 'x', 'y', or 'xy' (default).

        no_reset : bool
            Do not reset ``plot_autozoom`` setting.

        """
        # calculate actual width of the limits/image, considering rotation
        (x_min, y_min), (x_max, y_max) = self.get_limits()
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()

        # account for t_[scale_x/y_base]
        if axis in ['x', 'xy']:
            x_lo, x_hi = x_min, x_max
        if axis in ['y', 'xy']:
            y_lo, y_hi = y_min, y_max

        self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

        if not no_reset:
            if self.t_['plot_autozoom'] == 'once':
                self.t_.set(plot_autozoom='off')

    def set_dist_axis(self, x_axis=None, y_axis=None):
        if x_axis is None:
            x_axis = self.t_['plot_dist_axis'][0]
        if y_axis is None:
            y_axis = self.t_['plot_dist_axis'][1]

        self.t_.set(plot_dist_axis=(x_axis, y_axis))

    def axis_dist_change_cb(self, setting, value):
        x_axis, y_axis = value
        plotable = self.get_dataobj()
        if x_axis is not None:
            if isinstance(plotable, Plotable):
                plotable.set(x_axis_scale=x_axis)
            self.ax.set_xscale(x_axis)
        if y_axis is not None:
            if isinstance(plotable, Plotable):
                plotable.set(y_axis_scale=y_axis)
            self.ax.set_yscale(y_axis)

        self.redraw()

    def redraw_now(self, whence=0):
        try:
            time_start = time.time()

            self.figure.canvas.draw()

            time_done = time.time()
            time_elapsed = time_done - time_start
            self.logger.debug(
                "widget '%s' redraw elapsed=%.4f sec" % (
                    self.name, time_elapsed))

            self.make_callback('redraw', whence)

        except Exception as e:
            self.logger.error(f"Error redrawing image: {e}", exc_info=True)

    # Because we may implement deferred redrawing in the future
    redraw = redraw_now

    def render_canvas(self, canvas):
        for obj in canvas.objects:
            if isinstance(obj, self.dc.Path):
                data_x, data_y = obj.points.T
                self._plot_line(data_x, data_y, linewidth=obj.linewidth,
                                linestyle=obj.linestyle, color=obj.color,
                                alpha=obj.alpha, name=obj.name)

            elif isinstance(obj, self.dc.Line):
                data_x, data_y = [obj.x1, obj.x2], [obj.y1, obj.y2]
                self._plot_line(data_x, data_y, linewidth=obj.linewidth,
                                linestyle=obj.linestyle, color=obj.color,
                                alpha=obj.alpha, name=obj.name)

            elif isinstance(obj, self.dc.Text):
                text_clr = colors.resolve_color(obj.color, obj.alpha,
                                                format='tuple')
                bg_clr = None
                if obj.bgcolor is not None:
                    bg_clr = colors.resolve_color(obj.bgcolor, obj.bgalpha,
                                                  format='tuple')
                bd_clr = None
                if obj.bordercolor is not None:
                    bd_clr = colors.resolve_color(obj.bordercolor, obj.borderalpha,
                                                  format='tuple')
                self._plot_text(obj.x, obj.y, text=obj.text,
                                rot_deg=obj.rot_deg,
                                color=text_clr, bgcolor=bg_clr,
                                bordercolor=bd_clr,
                                borderlinewidth=obj.borderlinewidth,
                                borderpadding=obj.borderpadding,
                                font=obj.font, fontsize=obj.fontsize)

            elif isinstance(obj, self.dc.Canvas):
                self.render_canvas(obj)

            elif isinstance(obj, self.dc.XRange):
                self._plot_xrange(obj.x1, obj.x2, fillcolor=obj.fillcolor,
                                  fillalpha=obj.fillalpha)

            elif isinstance(obj, self.dc.YRange):
                self._plot_xrange(obj.y1, obj.y2, fillcolor=obj.fillcolor,
                                  fillalpha=obj.fillalpha)

    def apply_profile_or_settings(self, dataobj):
        """Apply a profile to the viewer.

        Parameters
        ----------
        dataobj : `~ginga.plot.Plotable`
            Image object.

        This function is used to initialize the viewer when a new data object
        is loaded.  Either the embedded profile settings or the default
        settings are applied as specified in the channel preferences.
        """
        # 1. copy the current viewer settings
        tmpprof = Settings.SettingGroup()
        self.t_.copy_settings(tmpprof, include_callbacks=False,
                              keylist=self.profile_keylist,
                              callback=False)

        # 2. reset selected items in the copy to default profile
        # if there is one
        keylist1 = set([])
        if self.default_viewer_profile is not None:
            dvp = self.default_viewer_profile
            if self.t_['viewer_restore_range'] and 'plot_range' in dvp:
                keylist1.add('plot_range')

            # is this really needed if we restore the range?
            if self.t_['viewer_restore_pan'] and 'plot_pan' in dvp:
                keylist1.add('plot_pan')

            # TODO
            # if self.t_['viewer_restore_cuts'] and 'cuts' in dvp:
            #     keylist1.add('cuts')

            # if self.t_['viewer_restore_distribution'] and 'color_algorithm' in dvp:
            #     keylist1.add('color_algorithm')

            # if self.t_['viewer_restore_color_map'] and 'color_map' in dvp:
            #     keylist1.update({'color_map', 'color_map_invert',
            #                      'color_map_rot_pct'})

            dvp.copy_settings(tmpprof, keylist=list(keylist1), callback=False)

        # 3. apply embedded profile selected items to the copy
        keylist2 = set([])
        profile = dataobj.get('profile', None)
        if profile is not None:
            # is this really needed if we restore the range?
            if self.t_['profile_use_pan'] and 'plot_pan' in profile:
                keylist2.add('plot_pan')

            if self.t_['profile_use_range'] and 'plot_range' in profile:
                keylist2.add('plot_range')

            # TODO
            # if self.t_['profile_use_cuts'] and 'cuts' in profile:
            #     keylist2.add('cuts')

            # if self.t_['profile_use_distribution'] and 'color_algorithm' in profile:
            #     keylist2.add('color_algorithm')

            # if self.t_['profile_use_color_map'] and 'color_map' in profile:
            #     keylist2.update({'color_map', 'color_map_invert',
            #                      'color_map_rot_pct'})

            profile.copy_settings(tmpprof, keylist=list(keylist2),
                                  callback=False)

        # 4. update our settings from the copy
        keylist = list(keylist1.union(keylist2))
        self.apply_profile(tmpprof, keylist=keylist)

        # 5. proceed with initialization that is not in the profile
        # initialize scale
        if self.t_['plot_autozoom'] != 'off' and 'plot_range' not in keylist2:
            self.logger.debug("auto zoom (%s)" % (self.t_['plot_autozoom']))
            self.zoom_fit()

        # TODO
        # # initialize cuts
        # if self.t_['autocuts'] != 'off' and 'cuts' not in keylist2:
        #     self.logger.debug("auto cuts (%s)" % (self.t_['autocuts']))
        #     self.auto_levels()

        # save the profile in the image
        if self.use_embedded_profile:
            self.checkpoint_profile()

    def show_plotable(self, plotable):
        canvas = plotable.get_canvas()
        self.render_canvas(canvas)

        self.set_titles(title=plotable['title'],
                        x_axis=plotable['x_axis_label'],
                        y_axis=plotable['y_axis_label'],
                        redraw=False)
        self.set_grid(plotable['grid'])
        self.set_legend(plotable['legend'])
        self.set_background(plotable['clr_bg'])

        self.apply_profile_or_settings(plotable)

        self.redraw_now()

    def replot(self):
        dataobj = self._dataobj
        if dataobj is None:
            return

        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()

        self.clear_data()

        if isinstance(dataobj, Plotable):
            self.show_plotable(dataobj)

        self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def _dataobj_modified_cb(self, dataobj):
        """Called when a dataobj we have looked at before is modified."""
        _dataobj = self.get_dataobj()
        if _dataobj is None or _dataobj is not dataobj:
            # nothing to do or object being modified is not the one
            # we are displaying
            return

        self.replot()

    def get_pan(self, coord='data'):
        """Get pan positions.

        Parameters
        ----------
        coord : {'data'}
            For compatibility with the image viewer.

        Returns
        -------
        positions : tuple
            X and Y positions, in that order.

        """
        if coord != 'data':
            raise ValueError("`coord` must be 'data' with this viewer")
        return self.t_['plot_pan']

    def set_pan(self, pan_x, pan_y, coord='data', no_reset=False):
        """Set pan position.

        Parameters
        ----------
        pan_x, pan_y : float
            Pan positions in X and Y.

        coord : {'data', 'wcs'}
            Indicates whether the given pan positions are in data or WCS space.

        no_reset : bool
            Do not reset ``autocenter`` setting.

        """
        self.t_.set(plot_pan=(pan_x, pan_y))

    def pan_change_cb(self, setting, value):
        pan_x, pan_y = value

        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        x_delta, y_delta = (x_hi - x_lo) * 0.5, (y_hi - y_lo) * 0.5
        x1, x2 = pan_x - x_delta, pan_x + x_delta
        y1, y2 = pan_y - y_delta, pan_y + y_delta

        self.set_ranges(x_range=(x1, x2), y_range=(y1, y2))

    def panset_xy(self, data_x, data_y, no_reset=False):
        """Similar to :meth:`set_pan`, except that input pan positions
        are always in data space.

        """
        self.set_pan(data_x, data_y, coord='data', no_reset=no_reset)

    def pan_delta(self, x_delta, y_delta):
        """Pan by a delta in X and Y specified in data points.

        Parameters
        ----------
        x_delta : float
            Delta in X

        y_delta : float
            Delta in Y

        """
        pan_x, pan_y = self.get_pan(coord='data')[:2]
        pan_x += x_delta
        pan_y += y_delta
        self.panset_xy(pan_x, pan_y)

    def panset_pct(self, pct_x, pct_y):
        """Similar to :meth:`set_pan`, except that pan positions
        are determined by multiplying data dimensions with the given
        scale factors, where 1 is 100%.

        """
        xy_mn, xy_mx = self.get_limits()
        pan_x, pan_y = self.get_pan()[:2]

        if pct_x is not None:
            pan_x = (xy_mn[0] + xy_mx[0]) * pct_x
        if pct_y is not None:
            pan_y = (xy_mn[1] + xy_mx[1]) * pct_y

        self.panset_xy(pan_x, pan_y)

    def calc_pan_pct(self, pad=0, min_pct=0.0, max_pct=0.9):
        """Calculate values for vertical/horizontal panning by percentages
        from the current pan position.  Mostly used by scrollbar callbacks.

        Parameters
        ----------
        pad : int (optional, defaults to 0)
            a padding amount in pixels to add to the limits when calculating

        min_pct : float (optional, range 0.0:1.0, defaults to 0.0)
        max_pct : float (optional, range 0.0:1.0, defaults to 0.9)

        Returns
        -------
        res : `~ginga.misc.Bunch.Bunch`
            calculation results, which include the following attributes:
            - rng_x : the range of X of the limits (including padding)
            - rng_y : the range of Y of the limits (including padding)
            - vis_x : the visually shown range of X in the viewer
            - vis_y : the visually shown range of Y in the viewer
            - thm_pct_x : the length of a X scrollbar arm as a ratio
            - thm_pct_y : the length of a Y scrollbar arm as a ratio
            - pan_pct_x : the pan position of X as a ratio
            - pan_pct_y : the pan position of Y as a ratio

        """
        limits = self.get_limits()
        pan_x, pan_y = self.get_pan()[:2]

        # calculate the corners of the entire image in unscaled cartesian
        mxwd, mxht = limits[1][:2]
        mxwd, mxht = mxwd + pad, mxht + pad
        mnwd, mnht = limits[0][:2]
        mnwd, mnht = mnwd - pad, mnht - pad

        arr = np.array([(mnwd, mnht), (mxwd, mnht),
                        (mxwd, mxht), (mnwd, mxht)],
                       dtype=float)
        x, y = arr.T
        x, y = x - pan_x, y - pan_y
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)

        # this is the range of X and Y of the entire image
        # in the viewer (unscaled)
        rng_x, rng_y = abs(x_max - x_min), abs(y_max - y_min)

        # this is the *visually shown* range of X and Y
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        vis_x, vis_y = x_hi - x_lo, y_hi - y_lo
        arr = np.array([(x_lo, y_lo), (x_hi, y_hi)])
        x, y = arr.T
        x, y = x - pan_x, y - pan_y

        # calculate the length of the slider arms as a ratio
        if rng_x == 0.0:
            pct_x = 1.0
            xthm_pct = 1.0
        else:
            pct_x = min(max(0.0, abs(x_min) / rng_x), 1.0)
            xthm_pct = max(min_pct, min(vis_x / rng_x, max_pct))

        if rng_y == 0.0:
            pct_y = 1.0
            ythm_pct = 1.0
        else:
            pct_y = min(max(0.0, abs(y_min) / rng_y), 1.0)
            ythm_pct = max(min_pct, min(vis_y / rng_y, max_pct))

        bnch = Bunch.Bunch(rng_x=rng_x, rng_y=rng_y, vis_x=vis_x, vis_y=vis_y,
                           thm_pct_x=xthm_pct, thm_pct_y=ythm_pct,
                           pan_pct_x=pct_x, pan_pct_y=pct_y)
        return bnch

    def pan_by_pct(self, pct_x, pct_y, pad=0):
        """Pan by a percentage of the data space. This method is designed
        to be called by scrollbar callbacks.

        Parameters
        ----------
        pct_x : float (range 0.0 : 1.0)
            Percentage in the X range to pan

        pct_y : float (range 0.0 : 1.0)
            Percentage in the Y range to pan

        pad : int (optional, defaults to 0)
            a padding amount in pixels to add to the limits when calculating

        """
        # Sanity check on inputs
        pct_x = np.clip(pct_x, 0.0, 1.0)
        pct_y = np.clip(pct_y, 0.0, 1.0)

        limits = self.get_limits()

        mxwd, mxht = limits[1][:2]
        mxwd, mxht = mxwd + pad, mxht + pad
        mnwd, mnht = limits[0][:2]
        mnwd, mnht = mnwd - pad, mnht - pad

        arr = np.array([(mnwd, mnht), (mxwd, mnht),
                        (mxwd, mxht), (mnwd, mxht)],
                       dtype=float)
        x, y = arr.T
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)

        crd_x = x_min + (pct_x * (x_max - x_min))
        crd_y = y_min + (pct_y * (y_max - y_min))

        pan_x, pan_y = crd_x, crd_y
        self.logger.debug("crd=%f,%f pan=%f,%f" % (
            crd_x, crd_y, pan_x, pan_y))

        self.panset_xy(pan_x, pan_y)

    def zoom_plot(self, pct_x, pct_y, redraw=True):
        """Zoom the plot, keeping the pan position."""
        set_range = False
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()
        pan_x, pan_y = self.get_pan()[:2]

        if pct_x is not None:
            xrng = x_hi - x_lo
            xinc = (xrng * 0.5) * pct_x
            x_lo, x_hi = pan_x - xinc, pan_x + xinc
            set_range = True

        if pct_y is not None:
            yrng = y_hi - y_lo
            yinc = (yrng * 0.5) * pct_y
            y_lo, y_hi = pan_y - yinc, pan_y + yinc
            set_range = True

        if set_range:
            self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def zoom_plot_at_cursor(self, cur_x, cur_y, pct_x, pct_y):
        """Zoom the plot, keeping the position under the cursor."""
        set_range = False
        (x_lo, x_hi), (y_lo, y_hi) = self.get_ranges()

        if pct_x is not None:
            x_lo, x_hi = (cur_x - (cur_x - x_lo) * pct_x,
                          cur_x + (x_hi - cur_x) * pct_x)
            set_range = True

        if pct_y is not None:
            y_lo, y_hi = (cur_y - (cur_y - y_lo) * pct_y,
                          cur_y + (y_hi - cur_y) * pct_y)
            set_range = True

        if set_range:
            self.set_ranges(x_range=(x_lo, x_hi), y_range=(y_lo, y_hi))

    def get_axes_size_in_px(self):
        bbox = self.ax.get_window_extent().transformed(self.figure.dpi_scale_trans.inverted())
        width, height = bbox.width, bbox.height
        width *= self.figure.dpi
        height *= self.figure.dpi
        return (width, height)

    def get_rgb_array(self):
        canvas = self.get_widget()
        # force a draw
        canvas.draw()
        return self.plot_w.get_rgb_array()

    def get_rgb_image_as_buffer(self, output=None, format='png',
                                quality=90):
        """Get the current image shown in the viewer, with any overlaid
        graphics, in a file IO-like object encoded as a bitmap graphics
        file.

        This can be overridden by subclasses.

        Parameters
        ----------
        output : a file IO-like object or None
            open python IO descriptor or None to have one created

        format : str
            A string defining the format to save the image.  Typically
            at least 'jpeg' and 'png' are supported. (default: 'png')

        quality: int
            The quality metric for saving lossy compressed formats.

        Returns
        -------
        buffer : file IO-like object
            This will be the one passed in, unless `output` is None
            in which case a BytesIO obejct is returned

        """
        obuf = output
        if obuf is None:
            obuf = BytesIO()

        fig_dpi = self.settings.get('plot_save_dpi', 100)
        fig = self.get_figure()
        fig.savefig(obuf, dpi=fig_dpi)

        return obuf

    def save_rgb_image_as_file(self, filepath, format='png', quality=90):
        """Save the current image shown in the viewer, with any overlaid
        graphics, in a file with the specified format and quality.
        This can be overridden by subclasses.

        Parameters
        ----------
        filepath : str
            path of the file to write

        format : str
            See :meth:`get_rgb_image_as_buffer`.

        quality: int
            See :meth:`get_rgb_image_as_buffer`.

        """
        plot_ext = self.settings.get('plot_save_suffix', f'.{format}')
        if not filepath.endswith(plot_ext):
            filepath = filepath + plot_ext

        fig_dpi = self.settings.get('plot_save_dpi', 100)
        fig = self.get_figure()
        fig.savefig(filepath, dpi=fig_dpi)

    def get_cursor(self, cname):
        """Get the cursor stored under the name.
        This can be overridden by subclasses, if necessary.

        Parameters
        ----------
        cname : str
            name of the cursor to return.

        """
        return self.cursor[cname]

    def define_cursor(self, cname, cursor):
        """Define a viewer cursor under a name.  Does not change the
        current cursor.

        Parameters
        ----------
        cname : str
            name of the cursor to define.

        cursor : object
            a cursor object in the back end's toolkit

        `cursor` is usually constructed from `make_cursor`.
        """
        self.cursor[cname] = cursor

    def switch_cursor(self, cname):
        """Switch the viewer's cursor to the one defined under a name.

        Parameters
        ----------
        cname : str
            name of the cursor to switch to.

        """
        cursor = self.get_cursor(cname)
        self.figure.canvas.set_cursor(cursor)

    def apply_profile(self, profile, keylist=None):
        """Apply a profile to the viewer.

        Parameters
        ----------
        profile : `~ginga.misc.Settings.SettingGroup`

        This function is used to initialize the viewer to a known state.
        The keylist, if given, will limit the items to be transferred
        from the profile to viewer settings, otherwise all items are
        copied.
        """
        profile.copy_settings(self.t_, keylist=keylist,
                              callback=True)

    def capture_profile(self, profile):
        self.t_.copy_settings(profile, keylist=self.profile_keylist,
                              callback=False)
        self.logger.debug("profile attributes set")

    def capture_default_viewer_profile(self):
        if self.default_viewer_profile is None:
            self.default_viewer_profile = Settings.SettingGroup()
        self.t_.copy_settings(self.default_viewer_profile,
                              keylist=self.profile_keylist,
                              callback=False)
        self.logger.info("captured default profile")

    def checkpoint_profile(self):
        profile = self.save_profile()
        if profile is None:
            # no dataobj in viewer
            return
        self.capture_profile(profile)
        return profile

    def save_profile(self, **params):
        """Save the given parameters into profile settings.

        Parameters
        ----------
        params : dict
            Keywords and values to be saved.

        """
        dataobj = self.get_dataobj()
        if dataobj is None:
            return

        profile = dataobj.save_profile(**params)
        return profile

    def _update_profile_cb(self, setting, value):
        key = setting.name
        if self.use_embedded_profile:
            kwargs = {key: value}
            self.save_profile(**kwargs)

    def set_foreground(self, fg):
        self.clr_fg = colors.resolve_color(fg)

    def set_background(self, bg):
        self.clr_bg = colors.resolve_color(bg)
        # record in plotable
        plotable = self.get_dataobj()
        if isinstance(plotable, Plotable):
            plotable.set(clr_bg=self.clr_bg)

        self.ax.set_facecolor(self.clr_bg)

    def onscreen_message(self, text, delay=None, redraw=True):
        if self._ost is not None:
            self._ost.remove()

        width, height = self.get_window_size()

        font = self.t_.get('onscreen_font', 'sans serif')
        font_size = self.t_.get('onscreen_font_size', None)
        if font_size is None:
            font_size = font_asst.calc_font_size(width)

        self._ost = self.figure.text(0.5, 0.5, text,
                                     transform=self.figure.transFigure,
                                     ha='center', va='center', fontsize=16,
                                     font='Sans', color=self.clr_fg,
                                     # NOTE: setting facecolor by
                                     # bbox dict  method doesn't work with
                                     # an RGB triplet
                                     bbox={'facecolor': 'black',
                                           'linewidth': 0, 'fill': True,
                                           'alpha': 1.0, 'pad': 6})
        if redraw:
            self.redraw()

        if delay is not None and self.timer_msg is not None:
            self.timer_msg.set(delay)

    def onscreen_message_off(self):
        return self.onscreen_message('')

    def set_last_data_xy(self, data_x, data_y):
        self.last_data_x = data_x
        self.last_data_y = data_y

    def get_last_data_xy(self):
        return (self.last_data_x, self.last_data_y)

    check_cursor_location = get_last_data_xy

    def pixel_info_cb(self, canvas, pt, viewer, settings):
        """Called if no canvas above us has handled the 'pixel-info' callback.

        Return a Bunch that contains information about the position at
        the cursor.
        """
        data_x, data_y = pt[:2]
        # return generic info
        info = Bunch.Bunch(itype='base', data_x=data_x, data_y=data_y,
                           x=data_x, y=data_y, value=None)
        return info

    def set_enter_focus(self, tf):
        self.t_['plot_enter_focus'] = tf


class PlotViewEvent(PlotEventMixin, Mixins.UIMixin, PlotViewBase):

    def __init__(self, logger=None, settings=None, figure=None):
        PlotViewBase.__init__(self, logger=logger, settings=settings,
                              figure=figure)
        Mixins.UIMixin.__init__(self)
        PlotEventMixin.__init__(self)


class CanvasView(PlotViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, settings=None, figure=None,
                 bindmap=None, bindings=None):
        PlotViewEvent.__init__(self, logger=logger, settings=settings,
                               figure=figure)

        #self.private_canvas.ui_set_active(True, viewer=self)
        self.ui_set_active(True, viewer=self)

        if bindmap is None:
            bindmap = CanvasView.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = CanvasView.bindingsClass(self.logger)
        self.set_bindings(bindings)

        # TODO: defaults?
        bindings.enable(pan=True, zoom=True)

        self._mode = None
        self.modetbl = dict(locked='L', softlock='S', held='H', oneshot='O')
        self.bindmap.add_callback('mode-set', self._mode_set_cb)

        # TODO
        bindings.set_mode(self, 'plot2d', mode_type='locked')

        # Needed for UIMixin to propagate events correctly
        #self.objects = [self.private_canvas]

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)

    def _mode_set_cb(self, bindmap, mode, mode_type):

        width, height = self.get_window_size()

        font = self.t_.get('onscreen_font', 'sans serif')
        font_size = self.t_.get('onscreen_font_size', None)
        if font_size is None:
            font_size = font_asst.calc_font_size(width * 0.33)
            # TEMP
            if self._mode is None:
                font_size = 12

        # show the little mode status window
        if self._mode is not None:
            self._mode.remove()
            self._mode = None
        if mode is not None and self.t_.get('plot_show_mode', False):
            mode_txt = f"{mode} [{self.modetbl[mode_type]}]"
            self._mode = self.figure.text(0.92, 0.02, mode_txt,
                                          transform=self.figure.transFigure,
                                          ha='center', va='center',
                                          font=font, fontsize=font_size,
                                          color=self.clr_fg,
                                          # NOTE: setting facecolor by
                                          # bbox dict  method doesn't work with
                                          # an RGB triplet
                                          bbox={'facecolor': 'black',
                                                'fill': True, 'alpha': 1.0,
                                                'linewidth': 0, 'pad': 6})
        self.redraw()

    # def set_canvas(self, canvas, private_canvas=None):
    #     super(CanvasView, self).set_canvas(canvas,
    #                                        private_canvas=private_canvas)

    #     self.objects[0] = self.private_canvas


PlotViewGw = CanvasView
